#from Cryptodome.Cipher import AES
import struct
from hashlib import sha256

from .AES import AESModeOfOperationCTR, AESModeOfOperationCBC, Counter
from .blockfeeder import Decrypter, Encrypter, PADDING_NONE

def aesCCMEncrypt(plaintext, aad, key, nonce, macLen = 16):
	blockSize = 16 # For AES...
	if macLen not in (4, 6, 8, 10, 12, 14, 16):
		raise ValueError("Parameter 'mac_len' must be even and in the range 4..16 (not %d)" % macLen)
	if not (nonce and 7 <= len(nonce) <= 13):
		raise ValueError("Length of parameter 'nonce' must be in the range 7..13 bytes")
			
	# Encryption
	q = 15 - len(nonce)
	#cipher = AES.new(key=key, mode=AES.MODE_CTR, nonce=struct.pack("B", q - 1) + nonce)
	t_nonce = struct.pack("B", q - 1) + nonce
	t_noncepad = b'\x00'*(blockSize - (len(t_nonce) % blockSize))
	t_nonce += t_noncepad
	#print(t_nonce.hex())
	ctr_start = Counter(int.from_bytes(t_nonce, byteorder='big', signed=False))
	cipher = AESModeOfOperationCTR(key, ctr_start)
	s0 = cipher.encrypt(b'\x00'*16) # For mac
	#print(f"My s0 {s0}")
	c =  cipher.encrypt(plaintext)
	
	# Mac
	pLen = len(plaintext)
	aadLen = len(aad)
	#mac = AES.new(key=key, mode=AES.MODE_CBC, iv=b'\x00'*blockSize)
	mac = Encrypter(AESModeOfOperationCBC(key, iv=b'\x00'*blockSize), padding = PADDING_NONE)
	flags = (64 * (aadLen > 0) + 8 * ((macLen - 2) // 2) + (q - 1))
	b0 = struct.pack("B", flags) + nonce +  pLen.to_bytes(q, 'big')
	
	
	assocLenEncoded = b''
	if aadLen > 0:
		if aadLen < (2 ** 16 - 2 ** 8):
			encSize = 2
		elif aadLen < (2 ** 32):
			assocLenEncoded = b'\xFF\xFE'
			encSize = 4
		else:
			assocLenEncoded = b'\xFF\xFF'
			encSize = 8
		
		assocLenEncoded += aadLen.to_bytes(encSize, 'big')
		
	
	aadPadded =  assocLenEncoded + aad 
	#print(f"aad Format before pad: {len(aadPadded)}" )
	aadPad = b''
	if len(aadPadded) % blockSize != 0:
		#print("need to padd aad")
		aadPad =  b'\x00'*(blockSize - (len(aadPadded) % blockSize))
	
	aadPadded += aadPad
	ptxtPadded = plaintext 
	#ptxt padding
	ptxtPad = b''
	if (pLen % blockSize) != 0:
		#print("Should pad ptxt")
		ptxtPad = b'\x00'*(blockSize - (pLen % blockSize))
	
	ptxtPadded += ptxtPad
	
	macData = b0  + aadPadded + ptxtPadded
	#print(f"MAC input {macData}")
	t = mac.feed(macData)
	t += mac.feed()
	t = t[-16:]
	tag = bytes([a ^ b for (a,b) in zip(t,s0)])[:macLen] 
	return (c, tag)



def aesCCMDecrypt(ciphertext, aad, key, nonce, macValue):
	# Decrytion: in CTR Encrypt == Decrypt
	blockSize = 16 # For AES...
	macLen = len(macValue)
	if macLen not in (4, 6, 8, 10, 12, 14, 16):
		raise ValueError("Parameter 'mac_len' must be even and in the range 4..16 (not %d)" % macLen)
	if not (nonce and 7 <= len(nonce) <= 13):
		 raise ValueError("Length of parameter 'nonce' must be in the range 7..13 bytes")
			
	# Decryption
	q = 15 - len(nonce) 
	#cipher = AES.new(key=key, mode=AES.MODE_CTR, nonce=struct.pack("B", q - 1) + nonce)
	t_nonce = struct.pack("B", q - 1) + nonce
	t_noncepad = b'\x00'*(blockSize - (len(t_nonce) % blockSize))
	t_nonce += t_noncepad
	#print(t_nonce.hex())
	ctr_start = Counter(int.from_bytes(t_nonce, byteorder='big', signed=False))
	cipher = AESModeOfOperationCTR(key, ctr_start)

	s0 = cipher.encrypt(b'\x00'*16) # For mac
	#print(f"My s0 {s0}")
	plaintext =  cipher.encrypt(ciphertext)
	#print(f"recoverd plaintex {plaintext}")
	# Mac
	pLen = len(plaintext)
	aadLen = len(aad)
	#mac = AES.new(key=key, mode=AES.MODE_CBC, iv=b'\x00'*blockSize)
	mac = Encrypter(AESModeOfOperationCBC(key, iv=b'\x00'*blockSize), padding = PADDING_NONE)
	flags = (64 * (aadLen > 0) + 8 * ((macLen - 2) // 2) + (q - 1))
	b0 = struct.pack("B", flags) + nonce +  pLen.to_bytes(q, 'big')
	
	
	assocLenEncoded = b''
	if aadLen > 0:
		if aadLen < (2 ** 16 - 2 ** 8):
			encSize = 2
		elif aadLen < (2 ** 32):
			assocLenEncoded = b'\xFF\xFE'
			encSize = 4
		else:
			assocLenEncoded = b'\xFF\xFF'
			encSize = 8
		
		assocLenEncoded += aadLen.to_bytes(encSize, 'big')
		
	
	aadPadded =  assocLenEncoded + aad 
	#print(f"aad Format before pad: {len(aadPadded)}" )
	aadPad = b''
	if len(aadPadded) % blockSize != 0:
		#print("need to padd aad")
		aadPad =  b'\x00'*(blockSize - (len(aadPadded) % blockSize))
	
	aadPadded += aadPad
	ptxtPadded = plaintext 
	#ptxt padding
	ptxtPad = b''
	if pLen % blockSize != 0:
		#print("Should pad ptxt")
		ptxtPad = b'\x00'*(blockSize - (pLen % blockSize))
	
	ptxtPadded += ptxtPad
	
	macData = b0  + aadPadded + ptxtPadded
	#t = mac.encrypt(macData)[-16:]
	t = mac.feed(macData)
	t += mac.feed()
	t = t[-16:]
	tag = bytes([a ^ b for (a,b) in zip(t,s0)])[:macLen] 
	
	return plaintext
	
	
	## Attempt to secure comparison... Idea: hash expected and received macs and compare the result in constant time...
	## Ideally should be done with HMAC with a RANDOM KEY! Doesn't cost much on a performance level.
	## For now we use shake_128(SHA3 without a key)
	print(f"received mac {macValue} \nComputed {tag}")
	h1 = sha256()
	h1.update(tag)
	digest1 = h1.digest() 
	
	h2 = sha256() 
	h2.update(macValue)
	digest2 = h2.digest()
	
	# Constant time comparison of hashes. Probably overkill here because of the randomization introduced by HMAC..
	result = 0
	for x, y in zip(digest1, digest2):
		result |= x ^ y
	
	#print(f"Reuslt {result}")
	
	if result != 0:
		raise ValueError("Incorrect MAC")
	else:
		return plaintext


import enum
import io
import datetime
import uuid

from aiosmb.commons.utils.ts2dt import *
from aiosmb.protocol.common import NegotiateDialects
from aiosmb.wintypes.dtyp.constrcuted_security.guid import *


# https://msdn.microsoft.com/en-us/library/cc246543.aspx
class NegotiateSecurityMode(enum.IntFlag):
	SMB2_NEGOTIATE_SIGNING_ENABLED  = 0x0001
	SMB2_NEGOTIATE_SIGNING_REQUIRED = 0x0002


# https://msdn.microsoft.com/en-us/library/cc246543.aspx
class NegotiateCapabilities(enum.IntFlag):
	DFS = 0x00000001 #When set, indicates that the client supports the Distributed File System (DFS).
	LEASING = 0x00000002 #When set, indicates that the client supports leasing.
	LARGE_MTU = 0x00000004 #When set, indicates that the client supports multi-credit operations.
	MULTI_CHANNEL = 0x00000008 #When set, indicates that the client supports establishing multiple channels for a single session.
	PERSISTENT_HANDLES = 0x00000010 #When set, indicates that the client supports persistent handles.
	DIRECTORY_LEASING = 0x00000020 #When set, indicates that the client supports directory leasing.
	ENCRYPTION = 0x00000040 #When set, indicates that the client supports encryption.


# https://msdn.microsoft.com/en-us/library/cc246543.aspx
class NEGOTIATE_REQ:
	def __init__(self):
		self.StructureSize   = 36
		self.DialectCount    = None
		self.SecurityMode    = None
		self.Reserved        = 0
		self.Capabilities    = None
		self.ClientGuid      = None			
		self.ClientStartTime = 0
		self.NegotiateContextOffset = None
		self.NegotiateContextCount = None
		self.Reserved2 = 0

		self.NegotiateContextList = []
		self.Dialects        = []
		
	def to_bytes(self):
		if self.Dialects == []:
			raise Exception('At least one dialect MUST be set!')

		t = self.StructureSize.to_bytes(2, byteorder='little', signed = False)
		t += len(self.Dialects).to_bytes(2, byteorder='little', signed = False)
		t += self.SecurityMode.to_bytes(2, byteorder='little', signed = False)
		t += self.Reserved.to_bytes(2, byteorder='little', signed = False)
		t += self.Capabilities.to_bytes(4, byteorder='little', signed = False)
		t += self.ClientGuid.to_bytes()
		if NegotiateDialects.SMB311 in self.Dialects:
			dp = b''
			for dialect in self.Dialects:
				dp += dialect.value.to_bytes(2, byteorder='little', signed = False)
			
			padding_starts = 64+2+2+2+2+4+16+8+len(dp)
			dp += b'\x00' * ((8 - padding_starts) % 8)
			self.NegotiateContextOffset = 64+2+2+2+2+4+16+8+len(dp)
			self.NegotiateContextCount = len(self.NegotiateContextList)
			t += self.NegotiateContextOffset.to_bytes(4, byteorder='little', signed = False)
			t += self.NegotiateContextCount.to_bytes(2, byteorder='little', signed = False)
			t += self.Reserved2.to_bytes(2, byteorder='little', signed = False)
			t += dp

			for ctx in self.NegotiateContextList:
				t += ctx.to_bytes()
				t += b'\x00' * ((8 - len(t)) % 8)
			
		else:
			t += self.ClientStartTime.to_bytes(8, byteorder='little', signed = False)
			
			for dialect in self.Dialects:
				t += dialect.value.to_bytes(2, byteorder='little', signed = False)
		
		return t

	@staticmethod
	def from_buffer(buff):
		cmd = NEGOTIATE_REQ()
		cmd.StructureSize = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		assert cmd.StructureSize == 36
		cmd.DialectCount = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		assert cmd.DialectCount > 0
		cmd.SecurityMode = NegotiateSecurityMode(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		cmd.Reserved = buff.read(2)
		cmd.Capabilities = NegotiateCapabilities(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		cmd.ClientGuid = GUID.from_buffer(buff)
		# skipping the next field because it's interpretation depends on the data after it...
		pos = buff.tell()
		buff.seek(8, io.SEEK_CUR)
		
		cmd.Dialects = []		
		for _ in range(0, cmd.DialectCount):
			cmd.Dialects.append(NegotiateDialects(int.from_bytes(buff.read(2), byteorder = 'little', signed = False)))

		pos_buff_end = buff.tell()
		buff.seek(pos, io.SEEK_SET)

		if NegotiateDialects.SMB311 in cmd.Dialects:
			cmd.NegotiateContextOffset = int.from_bytes(buff.read(4),byteorder = 'little', signed = False)
			cmd.NegotiateContextCount  = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
			cmd.Reserved2 = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
			cmd.NegotiateContextList = []
			buff.seek(cmd.NegotiateContextOffset, io.SEEK_SET)

			for _ in range(0, cmd.NegotiateContextCount):
				cmd.NegotiateContextList.append(SMB2NegotiateContext.from_buffer(buff))
				pad_pos = buff.tell()
				#aligning buffer, because the next data must be on 8-byte aligned position
				q,m = divmod(pad_pos, 8)
				if m != 0:
					buff.seek((q+1)*8, io.SEEK_SET)
		else:
			cmd.ClientStartTime = int.from_bytes(buff.read(8), byteorder = 'little', signed = False)
			buff.seek(pos_buff_end, io.SEEK_SET)

		return cmd

	def __repr__(self):
		t = '==== SMB2 NEGOTIATE REQ ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'DialectCount:  %s\r\n' % self.DialectCount
		t += 'SecurityMode:  %s\r\n' % self.SecurityMode.name
		t += 'Reserved:      %s\r\n' % self.Reserved
		t += 'Capabilities:  %s\r\n' % repr(self.Capabilities)
		t += 'ClientGuid:    %s\r\n' % self.ClientGuid
		if NegotiateDialects.SMB311 in self.Dialects:
			t += 'NegotiateContextOffset:    %s\r\n' % self.NegotiateContextOffset
			t += 'NegotiateContextCount:    %s\r\n' % self.NegotiateContextCount
			t += 'Reserved2:    %s\r\n' % self.Reserved2
			for ctx in self.NegotiateContextList:
				t += repr(ctx)
		else:
			t += 'ClientStartTime:    %s\r\n' % self.ClientStartTime
		
		for dialect in self.Dialects:
			t += '\t Dialect: %s\r\n' % dialect.name

		return t


class SMB2ContextType(enum.Enum):
	PREAUTH_INTEGRITY_CAPABILITIES = 0x0001
	ENCRYPTION_CAPABILITIES = 0x0002
	COMPRESSION_CAPABILITIES = 0x0003 #The Data field contains a list of compression algorithms, as specified in section 2.2.3.1.3<13>.
	NETNAME_NEGOTIATE_CONTEXT_ID = 0x0005 #The Data field contains the server name to which the client connects<14>.


class SMB2HashAlgorithm(enum.Enum):
	SHA_512 = 0x0001

#
# https://msdn.microsoft.com/en-us/library/mt208834.aspx
class SMB2NegotiateContext:
	def __init__(self):
		pass

	@staticmethod
	def from_buffer(buff):
		pos = buff.tell()
		ContextType = SMB2ContextType(int.from_bytes(buff.read(2), byteorder = 'little', signed = False))
		buff.seek(pos)

		if ContextType == SMB2ContextType.PREAUTH_INTEGRITY_CAPABILITIES:
			return SMB2PreauthIntegrityCapabilities.from_buffer(buff)

		elif ContextType == SMB2ContextType.ENCRYPTION_CAPABILITIES:
			return SMB2EncryptionCapabilities.from_buffer(buff)

		elif ContextType == SMB2ContextType.COMPRESSION_CAPABILITIES:
			return SMB2CompressionCapabilities.from_buffer(buff)

		elif ContextType == SMB2ContextType.PREAUTH_INTEGRITY_CAPABILITIES:
			return SMB2NetnameNegotiateContextID.from_buffer(buff)

		else:
			raise Exception('Unknown contextype %s' % ContextType)

class SMB2PreauthIntegrityCapabilities:
	def __init__(self):
		self.ContextType = SMB2ContextType.PREAUTH_INTEGRITY_CAPABILITIES
		self.DataLength  = None
		self.Reserved    = 0

		self.HashAlgorithmCount = None
		self.SaltLength         = None
		self.HashAlgorithms     = None
		self.Salt               = None

	@staticmethod
	def from_buffer(buff):
		cap = SMB2PreauthIntegrityCapabilities()
		cap.ContextType = SMB2ContextType(int.from_bytes(buff.read(2), byteorder = 'little', signed = False))
		cap.DataLength  = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
		cap.Reserved    = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		cap.HashAlgorithmCount = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cap.SaltLength = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
		cap.HashAlgorithms = []
		for _ in range(cap.HashAlgorithmCount):
			cap.HashAlgorithms.append(SMB2HashAlgorithm(int.from_bytes(buff.read(2), byteorder ='little', signed = False)))

		cap.Salt = buff.read(cap.SaltLength)

		return cap

	@staticmethod
	def construct(hash_algos = [SMB2HashAlgorithm.SHA_512], salt = b''):
		cap = SMB2PreauthIntegrityCapabilities()
		cap.HashAlgorithms = hash_algos
		cap.Salt = salt
		return cap

	def to_bytes(self):
		self.SaltLength = len(self.Salt)
		self.HashAlgorithmCount = len(self.HashAlgorithms)
		data = self.HashAlgorithmCount.to_bytes(2, byteorder = 'little', signed=False)
		data += self.SaltLength.to_bytes(2, byteorder = 'little', signed=False)
		for hashalgo in self.HashAlgorithms:
			data += hashalgo.value.to_bytes(2, byteorder = 'little', signed=False)
		data += self.Salt

		self.DataLength = len(data)

		t = self.ContextType.value.to_bytes(2, byteorder = 'little', signed=False)
		t += self.DataLength.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Reserved.to_bytes(4, byteorder = 'little', signed=False)
		t += data
		
		return t

	def __repr__(self):
		t = '==== SMB2 Preauth Integrity Capabilities ====\r\n'
		t += 'HashAlgorithmCount: %s\r\n' % self.HashAlgorithmCount
		t += 'SaltLength: %s\r\n' % self.SaltLength
		t += 'Salt: %s\r\n' % self.Salt
		
		for algo in self.HashAlgorithms:
			t += 'HashAlgo: %s\r\n' % algo.name

		return t


class SMB2Cipher(enum.Enum):
	AES_128_CCM = 0x0001
	AES_128_GCM = 0x0002


class SMB2EncryptionCapabilities:
	def __init__(self):
		self.ContextType = SMB2ContextType.ENCRYPTION_CAPABILITIES
		self.DataLength  = None
		self.Reserved    = 0

		self.CipherCount = None
		self.Ciphers = None

	@staticmethod
	def from_buffer(buff):
		cap = SMB2EncryptionCapabilities()
		cap.ContextType = SMB2ContextType(int.from_bytes(buff.read(2), byteorder = 'little', signed = False))
		cap.DataLength  = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
		cap.Reserved    = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		cap.CipherCount = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cap.Ciphers = []

		for _ in range(cap.CipherCount):
			cap.Ciphers.append(SMB2Cipher(int.from_bytes(buff.read(2), byteorder='little', signed = False)))

		return cap

	@staticmethod
	def from_enc_list(supp_enc_list):
		#supp_enc_list > a list of SMB2Cipher enums
		s = SMB2EncryptionCapabilities()
		s.Ciphers = supp_enc_list
		s.CipherCount = len(supp_enc_list)
		return s

	def to_bytes(self):
		data  = self.CipherCount.to_bytes(2, byteorder = 'little', signed=False)
		for cipher in self.Ciphers:
			data += cipher.value.to_bytes(2, byteorder = 'little', signed=False)

		self.DataLength = len(data)
		t = self.ContextType.value.to_bytes(2, byteorder = 'little', signed=False)
		t += self.DataLength.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Reserved.to_bytes(4, byteorder = 'little', signed=False)
		t += data


		return t
			

	def __repr__(self):
		t = '==== SMB2 Encryption Capabilities ====\r\n'
		t += 'CipherCount: %s\r\n' % self.CipherCount
		for cipher in self.Ciphers:
			t += 'Cipher: %s\r\n' % cipher.name

		return t

class SMB2CompressionType(enum.Enum):
	NONE = 0x0000 #No compression
	LZNT1 = 0x0001 #LZNT1 compression algorithm
	LZ77 = 0x0002 #LZ77 compression algorithm
	LZ77_Huffman = 0x0003 #LZ77+Huffman compression algorithm
	Pattern_V1 = 0x0004 #Pattern Scanning algorithm

class SMB2CompressionFlags(enum.Enum):
	NONE = 0x00000000 #Chained compression is not supported.
	CHAINED = 0x00000001 #Chained compression is supported on this connection.

class SMB2CompressionCapabilities:
	def __init__(self):
		self.ContextType = SMB2ContextType.COMPRESSION_CAPABILITIES
		self.DataLength  = None
		self.Reserved    = 0

		self.CompressionAlgorithmCount = None
		self.Padding = 0
		self.Flags = None
		self.CompressionAlgorithms = []

	@staticmethod
	def from_buffer(buff):
		cap = SMB2CompressionCapabilities()
		cap.ContextType = SMB2ContextType(int.from_bytes(buff.read(2), byteorder = 'little', signed = False))
		cap.DataLength  = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
		cap.Reserved    = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		cap.CompressionAlgorithmCount = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cap.Padding = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cap.Flags = SMB2CompressionFlags(int.from_bytes(buff.read(4), byteorder='little', signed = False))

		for _ in range(cap.CompressionAlgorithmCount):
			cap.CompressionAlgorithms.append(SMB2CompressionType(int.from_bytes(buff.read(2), byteorder='little', signed = False)))

		return cap

	@staticmethod
	def from_comp_list(supp_comp_list, is_chained = False):
		#supp_comp_list > a list of SMB2CompressionType enums
		# flags > bool, will be converted to SMB2CompressionFlags 
		s = SMB2CompressionCapabilities()
		s.Flags = SMB2CompressionFlags.NONE
		if is_chained is True:
			s.Flags = SMB2CompressionFlags.CHAINED
		s.CompressionAlgorithms = supp_comp_list
		s.CompressionAlgorithmCount = len(supp_comp_list)
		return s

	def to_bytes(self):
		
		data  = self.CompressionAlgorithmCount.to_bytes(2, byteorder = 'little', signed=False)
		data += self.Padding.to_bytes(2, byteorder = 'little', signed=False)
		data += self.Flags.value.to_bytes(2, byteorder = 'little', signed=False)
		data += self.CompressionAlgorithmCount.to_bytes(2, byteorder = 'little', signed=False)
		for compalgo in self.CompressionAlgorithms:
			data += compalgo.value.to_bytes(2, byteorder = 'little', signed=False)
		
		self.DataLength = len(data)
		t = self.ContextType.value.to_bytes(2, byteorder = 'little', signed=False)
		t += self.DataLength.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Reserved.to_bytes(4, byteorder = 'little', signed=False)
		t += data

		return t
			

	def __repr__(self):
		t = '==== SMB2 Compression Capabilities ====\r\n'
		t += 'Flags: %s\r\n' % self.Flags.name
		for compalgo in self.CompressionAlgorithms:
			t += 'Compression type: %s\r\n' % compalgo.name

		return t

class SMB2NetnameNegotiateContextID:
	def __init__(self):
		self.ContextType = SMB2ContextType.NETNAME_NEGOTIATE_CONTEXT_ID
		self.DataLength  = None
		self.Reserved    = None

		self.NetName = None

	@staticmethod
	def from_buffer(buff):
		#raise Exception('Documentation doesnt specify the size')
		cap = SMB2NetnameNegotiateContextID()
		cap.ContextType = SMB2ContextType(int.from_bytes(buff.read(2), byteorder = 'little', signed = False))
		cap.DataLength  = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
		cap.Reserved    = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		cap.NetName     = buff.read(cap.DataLength).decode('utf-16-le')
		return cap

	def to_bytes(self):
		return self.NetName.encode('utf-18-le')
			
	def __repr__(self):
		t = '==== SMB2 Netname Negotiate Context ID ====\r\n'
		t += 'NetName: %s\r\n' % self.NetName
		return t

# https://msdn.microsoft.com/en-us/library/cc246561.aspx
class NEGOTIATE_REPLY:
	def __init__(self):
		self.StructureSize = 65
		self.SecurityMode = None
		self.DialectRevision = None
		self.NegotiateContextCount = None # or reserved
		self.ServerGuid = None
		self.Capabilities = None
		self.MaxTransactSize = 8388608
		self.MaxReadSize = 8388608
		self.MaxWriteSize = 8388608
		self.SystemTime = None
		self.ServerStartTime = None
		self.SecurityBufferOffset = 0
		self.SecurityBufferLength = 0
		self.NegotiateContextOffset = 0
		self.Buffer = b''
		self.Padding = None
		self.NegotiateContextList = []

	@staticmethod
	def from_bytes(bbuff):
		return NEGOTIATE_REPLY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = NEGOTIATE_REPLY()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		assert msg.StructureSize == 65
		msg.SecurityMode    = NegotiateSecurityMode(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.DialectRevision = NegotiateDialects(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.NegotiateContextCount  = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.ServerGuid      = GUID.from_bytes(buff.read(16))
		msg.Capabilities    = NegotiateCapabilities(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		msg.MaxTransactSize = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.MaxReadSize     = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.MaxWriteSize    = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.SystemTime      = timestamp2datetime(buff.read(8))
		msg.ServerStartTime = timestamp2datetime(buff.read(8))
		msg.SecurityBufferOffset = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.SecurityBufferLength = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.NegotiateContextOffset = int.from_bytes(buff.read(4), byteorder='little', signed = False)

		pos = buff.tell()
		if msg.SecurityBufferLength != 0:
			buff.seek(msg.SecurityBufferOffset, io.SEEK_SET)
			msg.Buffer = buff.read(msg.SecurityBufferLength)

		pos_buff_end = buff.tell()

		if msg.DialectRevision == NegotiateDialects.SMB311:
			msg.NegotiateContextList = []
			buff.seek(msg.NegotiateContextOffset, io.SEEK_SET)
			for _ in range(msg.NegotiateContextCount):
				msg.NegotiateContextList.append(SMB2NegotiateContext.from_buffer(buff))
				padsize = ((8 - buff.tell()) % 8)
				buff.seek(padsize, io.SEEK_CUR)

		return msg

	def to_bytes(self):
		self.SecurityBufferOffset = 0x80
		self.SecurityBufferLength = len(self.Buffer)

		t  = self.StructureSize.to_bytes(2, byteorder = 'little', signed=False)
		t += self.SecurityMode.to_bytes(2, byteorder = 'little', signed=False)
		t += self.DialectRevision.value.to_bytes(2, byteorder = 'little', signed=False)
		t += self.NegotiateContextCount.to_bytes(2, byteorder = 'little', signed=False)
		t += self.ServerGuid.to_bytes()
		t += self.Capabilities.to_bytes(4, byteorder = 'little', signed=False)
		t += self.MaxTransactSize.to_bytes(4, byteorder = 'little', signed=False)
		t += self.MaxReadSize.to_bytes(4, byteorder = 'little', signed=False)
		t += self.MaxWriteSize.to_bytes(4, byteorder = 'little', signed=False)
		t += datetime2timestamp(self.SystemTime)
		t += datetime2timestamp(self.ServerStartTime)

		t += self.SecurityBufferOffset.to_bytes(2, byteorder = 'little', signed=False)
		t += self.SecurityBufferLength.to_bytes(2, byteorder = 'little', signed=False)
		t += self.NegotiateContextOffset.to_bytes(4, byteorder = 'little', signed=False)
		t += self.Buffer

		if self.NegotiateContextCount > 0:
			for ngctx in self.NegotiateContextList:
				t+= ngctx.to_bytes()
				#PADDING!
				q,m = divmod(len(t)+64,8)
				t+= b'\x00'*( (q+1)*8 -  len(t)   )

		return t

	def __repr__(self):
		t = '==== SMB2 NEGOTIATE REPLY ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'SecurityMode: %s\r\n' % repr(self.SecurityMode)
		t += 'DialectRevision: %s\r\n' % self.DialectRevision.name
		t += 'ServerGuid: %s\r\n' % self.ServerGuid
		t += 'Capabilities: %s\r\n' % repr(self.Capabilities)
		t += 'MaxTransactSize: %s\r\n' % self.MaxTransactSize
		t += 'MaxReadSize: %s\r\n' % self.MaxReadSize
		t += 'MaxWriteSize: %s\r\n' % self.MaxWriteSize
		t += 'SystemTime: %s\r\n' % self.SystemTime.isoformat()
		t += 'ServerStartTime: %s\r\n' % self.ServerStartTime.isoformat()
		t += 'SecurityBufferOffset: %s\r\n' % self.SecurityBufferOffset
		t += 'SecurityBufferLength: %s\r\n' % self.SecurityBufferLength
		t += 'Buffer: %s\r\n' % self.Buffer
		t += 'NegotiateContextList: %s\r\n' % self.NegotiateContextList
		return t




class SMBException(Exception):
	def __init__(self, message = '', ntstatus = None):
		super().__init__(message)
		self.ntstatus = ntstatus

	def pprint(self):
		return 'Error! Server responded with: %s' % self.ntstatus.name

class SMBConnectionNetworkTerminated(SMBException):
	pass

class SMBConnectionTimeoutException(SMBException):
	pass
	
class SMBConnectionRefusedException(SMBException):
	pass
	
class SMBUnsupportedDialectSelected(SMBException):
	pass

class SMBUnsupportedDialectSign(SMBException):
	pass
	
class SMBUnsupportedSMBVersion(SMBException):
	pass
	
class SMBKerberosPreauthFailed(SMBException):
	pass

class SMBAuthenticationFailed(SMBException):
	pass
	
class SMBGenericException(SMBException):
	pass
	
class SMBIncorrectShareName(SMBException):
	pass
	
class SMBCreateAccessDenied(SMBException):
	pass

class SMBPendingTimeout(SMBException):
	pass

class SMBPendingMaxRenewal(SMBException):
	pass


##### not SMBException from this point!
class SMBMachineException(Exception):
	def __init__(self, message = ''):
		super().__init__(message)

class SMBSocks5ConnectionError(Exception):
	def __init__(self, message = ''):
		super().__init__(message)



import enum
import platform

class SMBCredentialsSecretType(enum.Enum):
	NT = 'NT'
	PASSWORD = 'PASSWORD'
	AES = 'AES'
	RC4 = 'RC4'
	CCACHE = 'CCACHE'
	NONE = 'NONE'

class SMBAuthProtocol(enum.Enum):
	NTLM = 'NTLM'
	KERBEROS = 'KERBEROS'
	SSPI_KERBEROS = 'SSPI_KERBEROS'
	SSPI_NTLM = 'SSPI_NTLM'
	MULTIPLEXOR_NTLM = 'MULTIPLEXOR_NTLM'
	MULTIPLEXOR_KERBEROS = 'MULTIPLEXOR_KERBEROS'
	MULTIPLEXOR_SSL_NTLM = 'MULTIPLEXOR_SSL_NTLM'
	MULTIPLEXOR_SSL_KERBEROS = 'MULTIPLEXOR_SSL_KERBEROS'


class SMBCredential:
	def __init__(self, username = None, domain = None, secret = None, secret_type = None, authentication_type = None, settings = None, target = None):
		self.username = username
		self.domain = domain
		self.secret = secret
		self.secret_type = secret_type #SMBCredentialsSecretType
		self.target = target #for kerberos authentication
		
		self.authentication_type = authentication_type #kerberos or NTLM or ...
		self.settings = settings
			
		#domain/user/auth_type/secret_type:secret@target_ip_hostname_fqdn:target_port/dc_ip

	def __str__(self):
		t = '==== SMBCredential ====\r\n'
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
			
		return t
	
	@staticmethod
	def from_credential_string(s):
		"""
		Making users life more conveinent
		"""
		return SMBCredential.from_connection_string(s + '@')

	@staticmethod
	def from_connection_string(s):
		creds = SMBCredential()
		
		t, target = s.rsplit('@', 1)
		creds.domain, t = t.split('/', 1)
		creds.username, t = t.split('/', 1)
		if t.find('/') != -1:
			auth_type , t = t.split('/', 1)
			secret_type, creds.secret = t.split(':',1)
			creds.secret_type = SMBCredentialsSecretType(secret_type.upper())
		else:
			auth_type = t
			creds.secret_type = SMBCredentialsSecretType.NONE
		creds.authentication_type = SMBAuthProtocol(auth_type.upper().replace('-','_'))
		
		#sanity check
		if creds.secret_type == [SMBCredentialsSecretType.NT, SMBCredentialsSecretType.RC4]:
			try:
				bytes.fromhex(creds.secret)
				if len(creds.secret) != 32:
					raise Exception()
			except Exception as e:
				raise Exception('This is not a valid NT hash')
				
		elif creds.secret_type == SMBCredentialsSecretType.AES:
			try:
				bytes.fromhex(creds.secret)
				if len(creds.secret) != 32 or len(creds.secret) != 64:
					raise Exception()
			except Exception as e:
				raise Exception('This is not a valid NT hash')
				
		elif creds.secret_type == SMBCredentialsSecretType.CCACHE:
			try:
				with open(creds.secret, 'rb') as f:
					a = 1
			except Exception as e:
				raise Exception('Could not open CCACHE file!')
		
		
		if creds.authentication_type == SMBAuthProtocol.NTLM:
			if creds.secret_type not in [SMBCredentialsSecretType.NT, SMBCredentialsSecretType.RC4, SMBCredentialsSecretType.PASSWORD]:
				raise Exception('NTLM authentication requires either password or NT hash or RC4 key to be specified as secret!')
				
		elif creds.authentication_type == SMBAuthProtocol.KERBEROS:
			if creds.secret_type not in [SMBCredentialsSecretType.NT, SMBCredentialsSecretType.RC4, SMBCredentialsSecretType.PASSWORD, SMBCredentialsSecretType.AES, SMBCredentialsSecretType.CCACHE]:
				raise Exception('KERBEROS authentication requires either password or NT hash or RC4 key or AES key or CCACHE file to be specified as secret!')
		
		elif creds.authentication_type in [SMBAuthProtocol.SSPI_NTLM,  SMBAuthProtocol.SSPI_KERBEROS]:
			if platform.system() != 'Windows':
				raise Exception('SSPI authentication on works on windows!')
		
		return creds
		
class SMBKerberosCredential:
	def __init__(self):
		self.mode = 'CLIENT'
		self.connection = None #KerberosCredential
		self.target = None #KerberosTarget
		self.ksoc = None #KerberosSocketAIO
		self.ccred = None
		
class SMBKerberosSSPICredential:
	def __init__(self):
		self.mode = 'CLIENT'
		self.client = None
		self.password = None
		self.target  = None
		
class SMBNTLMSSPICredential:
	def __init__(self):
		self.mode = 'CLIENT'
		self.client = None
		self.passwrd = None
		
class SMBNTLMCredential:
	def __init__(self):
		self.username = None
		self.domain = ''
		self.password = None
		self.workstation = None
		self.is_guest = False
		self.nt_hash = None
		self.lm_hash = None

class SMBMultiplexorCredential:
	def __init__(self):
		self.mode = 'CLIENT'
		self.type = 'NTLM'
		self.username = '<CURRENT>'
		self.domain = '<CURRENT>'
		self.password = '<CURRENT>'
		self.target = None
		self.is_guest = False
		self.is_ssl = False
		self.mp_host = None
		self.mp_port = None
		self.mp_username = None
		self.mp_domain = None
		self.mp_password = None
		self.agent_id = None

	def get_url(self):
		url_temp = 'ws://%s:%s'
		if self.is_ssl is True:
			url_temp = 'wss://%s:%s'
		url = url_temp % (self.mp_host, self.mp_port)
		return url

	def parse_settings(self, settings):
		self.mp_host = settings['host'][0]
		self.mp_port = settings['port'][0]
		if self.mp_port is None:
			self.mp_port = '9999'
		if 'user' in settings:
			self.mp_username = settings.get('user')[0]
		if 'domain' in settings:
			self.mp_domain = settings.get('domain')[0]
		if 'password' in settings:
			self.mp_password = settings.get('password')[0]
		self.agent_id = settings['agentid'][0]

		
def test():
	s = 'TEST/victim/ntlm/nt:AAAAAAAA@10.10.10.2:445'
	creds = SMBCredential.from_connection_string(s)
	print(str(creds))
	
	s = 'TEST/victim/sspi@10.10.10.2:445/aaaa'
	creds = SMBCredential.from_connection_string(s)
	
	print(str(creds))
	
if __name__ == '__main__':
	test()


from aiosmb.dcerpc.v5.common.connection.smbdcefactory import SMBDCEFactory
from aiosmb.dcerpc.v5 import srvs
from aiosmb.dcerpc.v5.dtypes import RPC_SID
from aiosmb.wintypes.ntstatus import NTStatus
from aiosmb import logger
from aiosmb.dcerpc.v5.dtypes import NULL
from aiosmb.commons.utils.decorators import red, rr, red_gen
		
class SMBSRVS:
	def __init__(self, connection):
		self.connection = connection
		self.service_manager = None
		
		self.dce = None
		self.handle = None
		
		self.domain_ids = {} #sid to RPC_SID
		self.domain_handles = {} #handle to sid
		
	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True,None
	
	@red
	async def connect(self, open = True):
		rpctransport = SMBDCEFactory(self.connection, filename=r'\srvsvc')
		self.dce = rpctransport.get_dce_rpc()
		await rr(self.dce.connect())
		await rr(self.dce.bind(srvs.MSRPC_UUID_SRVS))

		return True,None
	
	@red
	async def close(self):
		if self.dce:
			try:
				await self.dce.disconnect()
			except:
				pass
			return
		
		return True,None
	
	@red_gen
	async def list_shares(self, level = 1):
		level_name = 'Level%s' % level
		status = NTStatus.MORE_ENTRIES
		resumeHandle = 0
		while status == NTStatus.MORE_ENTRIES:
			resp, err = await srvs.hNetrShareEnum(self.dce, level, resumeHandle = resumeHandle)
			if err is not None:
				if err.error_code != NTStatus.MORE_ENTRIES.value:
					raise err
				resp = err.get_packet()

			for entry in resp['InfoStruct']['ShareInfo'][level_name]['Buffer']:
				yield entry['shi1_netname'][:-1], entry['shi1_type'], entry['shi1_remark'], None
			
			resumeHandle = resp['ResumeHandle'] 
			status = NTStatus(resp['ErrorCode'])	
	
	@red_gen
	async def list_sessions(self, level = 10):
		if level not in [1, 10]:
			raise Exception('Only levels 1 and 10 implemented!')
		level_name = 'Level%s' % level
		status = NTStatus.MORE_ENTRIES
		resumeHandle = 0
		while status == NTStatus.MORE_ENTRIES:
			resp, err = await srvs.hNetrSessionEnum(self.dce, '\x00', NULL, level, resumeHandle = resumeHandle)
			if err is not None:
				if err.error_code != NTStatus.MORE_ENTRIES.value:
					raise err
				resp = err.get_packet()

			if level == 1:
				for entry in resp['InfoStruct']['SessionInfo'][level_name]['Buffer']:
					username = entry['sesi1_username'][:-1]
					ip_addr = entry['sesi1_cname'][:-1]					
					yield username, ip_addr, None

			elif level == 10:
				for entry in resp['InfoStruct']['SessionInfo'][level_name]['Buffer']:
					username = entry['sesi10_username'][:-1]
					ip_addr = entry['sesi10_cname'][:-1]
					
					yield username, ip_addr, None
			
			resumeHandle = resp['ResumeHandle'] 
			status = NTStatus(resp['ErrorCode'])	



import re
## https://stackoverflow.com/questions/27726545/python-glob-but-against-a-list-of-strings-rather-than-the-filesystem
def glob2re(pat):
    """Translate a shell PATTERN to a regular expression.

    There is no way to quote meta-characters.
    """

    i, n = 0, len(pat)
    res = ''
    while i < n:
        c = pat[i]
        i = i+1
        if c == '*':
            #res = res + '.*'
            res = res + '[^/]*'
        elif c == '?':
            #res = res + '.'
            res = res + '[^/]'
        elif c == '[':
            j = i
            if j < n and pat[j] == '!':
                j = j+1
            if j < n and pat[j] == ']':
                j = j+1
            while j < n and pat[j] != ']':
                j = j+1
            if j >= n:
                res = res + '\\['
            else:
                stuff = pat[i:j].replace('\\','\\\\')
                i = j+1
                if stuff[0] == '!':
                    stuff = '^' + stuff[1:]
                elif stuff[0] == '^':
                    stuff = '\\' + stuff
                res = '%s[%s]' % (res, stuff)
        else:
            res = res + re.escape(c)
    return res + '\Z(?ms)'



# https://gist.github.com/ImmortalPC/c340564823f283fe530b
def hexdump(src, length=16, sep='.'):
	"""
	Pretty printing binary data blobs
	:param src: Binary blob
	:type src: bytearray
	:param length: Size of data in each row
	:type length: int
	:param sep: Character to print when data byte is non-printable ASCII
	:type sep: str(char)
	:return: str
	"""
	result = []

	for i in range(0, len(src), length):
		subSrc = src[i:i+length]
		hexa = ''
		isMiddle = False
		for h in range(0,len(subSrc)):
			if h == length/2:
				hexa += ' '
			h = subSrc[h]
			if not isinstance(h, int):
				h = ord(h)
			h = hex(h).replace('0x', '')
			if len(h) == 1:
				h = '0'+h
			hexa += h+' '
		hexa = hexa.strip(' ')
		text = ''
		for c in subSrc:
			if not isinstance(c, int):
				c = ord(c)
			if 0x20 <= c < 0x7F:
				text += chr(c)
			else:
				text += sep
		result.append(('%08X:  %-'+str(length*(2+1)+1)+'s  |%s|') % (i, hexa, text))

	return '\n'.join(result)


import enum
import asyncio
import hmac
import hashlib
import platform
import copy
import datetime

from aiosmb import logger
from aiosmb.exceptions import *
from aiosmb.network.network import TCPSocket
from aiosmb.network.netbios_transport import NetBIOSTransport
from aiosmb.protocol.smb.command_codes import SMBCommand
from aiosmb.commons.ntstatus import NTStatus
from aiosmb.protocol.smb.header import SMBHeader, SMBHeaderFlags2Enum
from aiosmb.protocol.smb.message import SMBMessage
from aiosmb.protocol.smb.commands import *
from aiosmb.protocol.smb2.message import SMB2Message, SMB2Transform
from aiosmb.protocol.smb2.commands import *
from aiosmb.protocol.smb2.headers import *
from aiosmb.protocol.smb2.command_codes import *
from aiosmb.protocol.common import *
from aiosmb.dtyp.constrcuted_security.guid import *
from aiosmb.commons.access_mask import *
from aiosmb.fscc.structures.fileinfoclass import *
from aiosmb.fscc.structures.FileFullDirectoryInformation import *
from aiosmb.fscc.FileAttributes import FileAttributes

from aiosmb.dtyp.constrcuted_security.security_descriptor import SECURITY_DESCRIPTOR

from aiosmb.commons.smbcontainer import *
from aiosmb.commons.smbtarget import *



class SMBConnectionStatus(enum.Enum):
	NEGOTIATING = 'NEGOTIATING'
	SESSIONSETUP = 'SESSIONSETUP'
	RUNNING = 'RUNNING'
	CLOSED = 'CLOSED'
	
class TreeEntry:
	def __init__(self):
		self.share_name = None
		self.tree_id = None
		self.session_id = None
		self.number_of_users = None
		self.is_DFS = None
		self.is_CA = None
		self.is_scaleout = None
		self.encrypt = None
		self.maximal_access = None
		
	@staticmethod
	def from_tree_reply(reply, share_name):
		te = TreeEntry()
		te.share_name = share_name
		te.tree_id = reply.header.TreeId
		te.session_id = reply.header.SessionId
		te.number_of_users = 1
		te.is_DFS = TreeCapabilities.SMB2_SHARE_CAP_DFS in reply.command.Capabilities
		te.is_CA = TreeCapabilities.SMB2_SHARE_CAP_CONTINUOUS_AVAILABILITY in reply.command.Capabilities
		te.is_scaleout = TreeCapabilities.SMB2_SHARE_CAP_SCALEOUT in reply.command.Capabilities
		te.encrypt = ShareFlags.SMB2_SHAREFLAG_ENCRYPT_DATA in reply.command.ShareFlags
		te.maximal_access = reply.command.MaximalAccess 
		return te
	
class FileHandle:
	def __init__(self):
		self.file_id = None
		self.tree_id = None
		self.oplock_level = None
		self.is_durable = None
		self.is_resilient = None
		self.last_disconnect_time = None
		self.file_name = None
	
	@staticmethod	
	def from_create_reply(reply, tree_id, file_name, oplock_level):
		fh = FileHandle()
		fh.file_id = reply.command.FileId
		fh.tree_id = tree_id
		fh.oplock_level = oplock_level
		fh.is_durable = False
		fh.is_resilient = False
		fh.last_disconnect_time = 0
		fh.file_name = file_name
		return fh

class SMBServerConnection:
	"""
	Connection class for network connectivity and SMB messages management (sending/recieveing/singing/encrypting).
	"""
	def __init__(self, server_settings, netbios_transport, dialects = [NegotiateDialects.SMB202], shutdown_evt = asyncio.Event()):
		self.settings = server_settings
		self.client_gssapi = self.settings.client_gssapi
		self.original_client_gssapi = copy.deepcopy(self.client_gssapi) #preserving a copy of the original
		self.shutdown_evt = shutdown_evt
		
		#######DONT CHANGE THIS
		self.supported_dialects = [NegotiateDialects.WILDCARD, NegotiateDialects.SMB202, NegotiateDialects.SMB210]
		#######
		
		self.settings = None
		self.network_transport = None 
		self.netbios_transport = netbios_transport
		self.dialects = dialects #list of SMBDialect
		
		self.selected_dialect = None
		self.signing_required = False
		self.encryption_required = False
		
		self.status = SMBConnectionStatus.NEGOTIATING
		
		self.OutstandingResponsesEvent = {}
		self.OutstandingRequests = {}
		self.OutstandingResponses = {}
		
		#two dicts for the same data, but with different lookup key
		self.TreeConnectTable_id = {}
		self.TreeConnectTable_share = {}
		
		self.FileHandleTable = {}
		
		self.SequenceWindow = 0
		self.MaxTransactSize = 0
		self.MaxReadSize = 0
		self.MaxWriteSize = 0
		self.ServerGuid = GUID.random()
		self.RequireSigning = False
		self.ServerName = None
		self.ClientGUID = None
		
		self.Dialect = 0
		self.SupportsFileLeasing = False
		self.SupportsMultiCredit = False
		
		self.SupportsDirectoryLeasing = False
		self.SupportsMultiChannel = False
		self.SupportsPersistentHandles = False
		self.SupportsEncryption = False
		self.ClientCapabilities = 0
		self.ServerCapabilities = 0
		self.ClientSecurityMode = 0
		self.ServerSecurityMode = 0
		
		
		self.SessionId = 0
		self.SessionKey = None
		
		
	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		await self.terminate()
		
	async def main(self):
		logger.info('SMB Server starting')
		logger.info('SMB Server creating NB transport reader task')
		asyncio.ensure_future(self.__handle_smb_in())
		
		logger.info('SMB Server running')
		while True:
			if self.status == SMBConnectionStatus.NEGOTIATING:
				await self.negotiate()
			if self.status == SMBConnectionStatus.SESSIONSETUP:
				await self.session_setup()
			else:
				raise Exception('Not implemented!')
		
	async def __handle_smb_in(self):
		"""
		Waits from SMB messages from the NetBIOSTransport in_queue, and fills the connection table.
		This function started automatically when calling connect.
		"""
		while not self.shutdown_evt.is_set():
			msg = await self.netbios_transport.in_queue.get()
			
			logger.log(1, '__handle_smb_in got new message with Id %s' % msg.header.MessageId)
			
			if isinstance(msg, SMB2Transform):
				#message is encrypted
				#this point we should decrypt it and only store the decrypted part in the OutstandingResponses table
				#but for now we just thropw exception bc encryption is not implemented
				raise Exception('Encrypted SMBv2 message recieved, but encryption is not yet supported!')
			
			self.OutstandingResponses[msg.header.MessageId] = msg
			if msg.header.MessageId in self.OutstandingResponsesEvent:
				self.OutstandingResponsesEvent[msg.header.MessageId].set()
			else:
				#here we are loosing messages, the functionality for "PENDING" and "SHARING_VIOLATION" should be written
				continue
		
		
	async def recvSMB(self, message_id):
		"""
		Returns an SMB message from the outstandingresponse dict, OR waits until the expected message_id appears.
		"""
		if message_id not in self.OutstandingResponses:
			self.OutstandingResponsesEvent[message_id] = asyncio.Event()
			await self.OutstandingResponsesEvent[message_id].wait()
		
		msg = self.OutstandingResponses.pop(message_id)
		
		return msg
	
		
	async def disconnect(self):
		"""
		Teras down the socket connecting as well as the reading cycle.
		Doesn't do any cleanup! 
		For proper cleanup call the terminate function.
		"""
		if self.status == SMBConnectionStatus.CLOSED:
			return
		
		self.status = SMBConnectionStatus.CLOSED
		self.shutdown_evt.set()
		await self.netbios_transport.stop()
		
		
		
	async def negotiate(self):
		"""
		"""

		rply = await self.recvSMB(0)

		#TODO: check if SMB2 is supported
		#currently we just continue with SMB2

		command = NEGOTIATE_REPLY()
		command.SecurityMode = NegotiateSecurityMode.SMB2_NEGOTIATE_SIGNING_ENABLED
		command.DialectRevision = NegotiateDialects.WILDCARD
		command.NegotiateContextCount = 0
		command.ServerGuid = self.ServerGuid
		command.Capabilities = 0
		command.SystemTime = datetime.datetime.now()
		command.ServerStartTime = datetime.datetime.now() - datetime.timedelta(days=1)
		command.Buffer = self.client_gssapi.get_mechtypes_list()

			
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.NEGOTIATE
		header.CreditReq = 0
		
		msg = SMBMessage(header, command)
		message_id = await self.sendSMB(msg)
		print(message_id)
		rply = await self.recvSMB(1)
		#recieveing reply, should be version2, because currently we dont support v1 :(
		 #negotiate MessageId should be 1
		
		print('1111111111111111111111111')

		#TODO: check if SMB2 is supported
		#currently we just continue with SMB2

		command = NEGOTIATE_REPLY()
		command.SecurityMode = NegotiateSecurityMode.SMB2_NEGOTIATE_SIGNING_ENABLED
		command.DialectRevision = NegotiateDialects.SMB202
		command.NegotiateContextCount = 0
		command.ServerGuid = self.ServerGuid
		command.Capabilities = 0
		command.SystemTime = datetime.datetime.now()
		command.ServerStartTime = datetime.datetime.now() - datetime.timedelta(days=1)
		command.Buffer = self.client_gssapi.get_mechtypes_list()

			
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.NEGOTIATE
		header.CreditReq = 0
		
		msg = SMBMessage(header, command)
		message_id = await self.sendSMB(msg)
		
		self.status = SMBConnectionStatus.SESSIONSETUP
		return

	async def session_setup(self):
		print('session_setup')
		rply = await self.recvSMB(2)
		self.SessionId = int.from_bytes(os.urandom(8), 'big', signed = False)
		auth_data = rply.command.Buffer
		print('clinet buffer: %s' % auth_data)
		data, res = await self.client_gssapi.authenticate(auth_data)


		command = SESSION_SETUP_REPLY()
		command.SessionFlags = 0
		command.Buffer = data

			
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.SESSION_SETUP
		header.CreditReq = 127
		header.Status = NTStatus.MORE_PROCESSING_REQUIRED
		
		msg = SMBMessage(header, command)
		message_id = await self.sendSMB(msg)
		print(message_id)
		rply = await self.recvSMB(3)


		
		
	def sign_message(self, msg):
		if self.selected_dialect in [NegotiateDialects.SMB202, NegotiateDialects.SMB210]:
			if self.SessionKey:
				msg.header.Flags = msg.header.Flags ^ SMB2HeaderFlag.SMB2_FLAGS_SIGNED ##maybe move this flag settings to sendsmb since singing is determined there?
				digest = hmac.new(self.SessionKey, msg.to_bytes(), hashlib.sha256).digest()
				msg.header.Signature = digest[:16]
		else:
			raise SMBUnsupportedDialectSign()
		
		
	async def sendSMB(self, msg):
		"""
		Sends an SMB message to teh remote endpoint.
		msg: SMB2Message or SMBMessage
		Returns: MessageId integer
		"""
		msg.header.Flags |= SMB2HeaderFlag.SMB2_FLAGS_SERVER_TO_REDIR
		if self.status == SMBConnectionStatus.NEGOTIATING:
			msg.header.CreditCharge = 1
			msg.header.CreditReq = 1
			msg.header.MessageId = self.SequenceWindow
			message_id = self.SequenceWindow
			self.SequenceWindow += 1
			
			self.OutstandingResponsesEvent[message_id] = asyncio.Event()
			await self.netbios_transport.out_queue.put(msg)
			return message_id
				

		if msg.header.Command is not SMB2Command.CANCEL:
			msg.header.MessageId = self.SequenceWindow
			self.SequenceWindow += 1
		
		msg.header.SessionId = self.SessionId
		
		if not msg.header.CreditCharge:
			msg.header.CreditCharge = 1
		
		if self.status != SMBConnectionStatus.SESSIONSETUP:
			msg.header.CreditReq = 127
		
		message_id = msg.header.MessageId
		
		if self.signing_required == True:
			self.sign_message(msg)
		
		if self.encryption_required == True:
			raise Exception('SMB Encryption not implemented yet :(')
			#self.encrypt_message(msg)
		
		###creating an event for outstanding response
		###self.OutstandingResponsesEvent[message_id] = asyncio.Event()

		if msg.header.Status != NTStatus.PENDING:
			if message_id in self.OutstandingResponsesEvent:
				del self.OutstandingResponsesEvent[message_id]
		else:
			self.OutstandingResponsesEvent[message_id].clear()
		
		await self.netbios_transport.out_queue.put(msg)
		
		return message_id
	
	
	
			
if __name__ == '__main__':
	target = SMBTarget()
	target.ip = '10.10.10.2'
	target.port = 445
	
	target_bad = SMBTarget()
	target_bad.ip = '10.10.10.66'
	target_bad.port = 445

	#asyncio.run(test(target))
	#asyncio.run(test_kerberos(target))
	#asyncio.run(test_sspi_kerberos(target))
	#asyncio.run(test_sspi_ntlm(target))
	#asyncio.run(connection_test(target))
	#asyncio.run(test_high(target))
	asyncio.run(filereader_test(target))
	
	


import io
import enum

from aiosmb.wintypes.fscc.FileAttributes import FileAttributes

class OplockLevel(enum.Enum):
	SMB2_OPLOCK_LEVEL_NONE = 0x00 #No oplock is requested.
	SMB2_OPLOCK_LEVEL_II = 0x01 #A level II oplock is requested.
	SMB2_OPLOCK_LEVEL_EXCLUSIVE = 0x08 #An exclusive oplock is requested.
	SMB2_OPLOCK_LEVEL_BATCH = 0x09 #A batch oplock is requested.
	SMB2_OPLOCK_LEVEL_LEASE = 0xFF #A lease is requested. If set, the request packet MUST contain an SMB2_CREATE_REQUEST_LEASE (section 2.2.13.2.8) create conte

class ImpersonationLevel(enum.Enum):
	Anonymous = 0x00000000 #The application-requested impersonation level is Anonymous.
	Identification = 0x00000001 #The application-requested impersonation level is Identification.
	Impersonation = 0x00000002 # The application-requested impersonation level is Impersonation.
	Delegate = 0x00000003 #The application-requested impersonation level is Delegate.

class ShareAccess(enum.IntFlag):
	FILE_SHARE_READ = 0x00000001 # When set, indicates that other opens are allowed to read this file while this open is present. This bit MUST NOT be set for a named pipe or a printer file. Each open creates a new instance of a named pipe. Likewise, opening a printer file always creates a new file.
	FILE_SHARE_WRITE = 0x00000002 # When set, indicates that other opens are allowed to write this file while this open is present. This bit MUST NOT be set for a named pipe or a printer file. Each open creates a new instance of a named pipe. Likewise, opening a printer file always creates a new file.
	FILE_SHARE_DELETE = 0x00000004 #When set, indicates that other opens are allowed to delete or rename this file while this open is present. This bit MUST NOT be set for a named pipe or a printer file. Each open creates a new instance of a named pipe. Likewise, opening a printer file always creates a new file.

class CreateDisposition(enum.Enum):
	FILE_SUPERSEDE = 0x00000000 #If the file already exists, supersede it. Otherwise, create the file. This value SHOULD NOT be used for a printer object.<30>
	FILE_OPEN = 0x00000001 #If the file already exists, return success; otherwise, fail the operation. MUST NOT be used for a printer object.
	FILE_CREATE = 0x00000002 #If the file already exists, fail the operation; otherwise, create the file.
	FILE_OPEN_IF = 0x00000003 #Open the file if it already exists; otherwise, create the file. This value SHOULD NOT be used for a printer object.<31>
	FILE_OVERWRITE = 0x00000004 #Overwrite the file if it already exists; otherwise, fail the operation. MUST NOT be used for a printer object.
	FILE_OVERWRITE_IF = 0x00000005 #Overwrite the file if it already exists; otherwise, create the file. This value SHOULD NOT be used for a printer object.<32>

class CreateOptions(enum.IntFlag):
	FILE_DIRECTORY_FILE = 0x00000001 #The file being created or opened is a directory file. With this flag, the CreateDisposition field MUST be set to FILE_CREATE, FILE_OPEN_IF, or FILE_OPEN. With this flag, only the following CreateOptions values are valid: FILE_WRITE_THROUGH, FILE_OPEN_FOR_BACKUP_INTENT, FILE_DELETE_ON_CLOSE, and FILE_OPEN_REPARSE_POINT. If the file being created or opened already exists and is not a directory file and FILE_CREATE is specified in the CreateDisposition field, then the server MUST fail the request with STATUS_OBJECT_NAME_COLLISION. If the file being created or opened already exists and is not a directory file and FILE_CREATE is not specified in the CreateDisposition field, then the server MUST fail the request with STATUS_NOT_A_DIRECTORY. The server MUST fail an invalid CreateDisposition field or an invalid combination of CreateOptions flags with STATUS_INVALID_PARAMETER.
	FILE_WRITE_THROUGH = 0x00000002 #The server performs file write-through; file data is written to the underlying storage before completing the write operation on this open.
	FILE_SEQUENTIAL_ONLY = 0x00000004 #This indicates that the application intends to read or write at sequential offsets using this handle, so the server SHOULD optimize for sequential access. However, the server MUST accept any access pattern. This flag value is incompatible with the FILE_RANDOM_ACCESS value.
	FILE_NO_INTERMEDIATE_BUFFERING = 0x00000008 #File buffering is not performed on this open; file data is not retained in memory upon writing it to, or reading it from, the underlying storage.
	FILE_SYNCHRONOUS_IO_ALERT = 0x00000010 #This bit SHOULD be set to 0 and MUST be ignored by the server.<34>
	FILE_SYNCHRONOUS_IO_NONALERT = 0x00000020 #This bit SHOULD be set to 0 and MUST be ignored by the server.<35>
	FILE_NON_DIRECTORY_FILE = 0x00000040 #If the name of the file being created or opened matches with an existing directory file, the server MUST fail the request with STATUS_FILE_IS_A_DIRECTORY. This flag MUST NOT be used with FILE_DIRECTORY_FILE or the server MUST fail the request with STATUS_INVALID_PARAMETER.
	FILE_COMPLETE_IF_OPLOCKED = 0x00000100 #This bit SHOULD be set to 0 and MUST be ignored by the server.<36>
	FILE_NO_EA_KNOWLEDGE = 0x00000200 #The caller does not understand how to handle extended attributes. If the request includes an SMB2_CREATE_EA_BUFFER create context, then the server MUST fail this request with STATUS_ACCESS_DENIED. If extended attributes with the FILE_NEED_EA flag (see [MS-FSCC] section 2.4.15) set are associated with the file being opened, then the server MUST fail this request with STATUS_ACCESS_DENIED.
	FILE_RANDOM_ACCESS = 0x00000800 #This indicates that the application intends to read or write at random offsets using this handle, so the server SHOULD optimize for random access. However, the server MUST accept any access pattern. This flag value is incompatible with the FILE_SEQUENTIAL_ONLY value. If both FILE_RANDOM_ACCESS and FILE_SEQUENTIAL_ONLY are set, then FILE_SEQUENTIAL_ONLY is ignored.
	FILE_DELETE_ON_CLOSE = 0x00001000 #The file MUST be automatically deleted when the last open request on this file is closed. When this option is set, the DesiredAccess field MUST include the DELETE flag. This option is often used for temporary files.
	FILE_OPEN_BY_FILE_ID = 0x00002000 #This bit SHOULD be set to 0 and the server MUST fail the request with a STATUS_NOT_SUPPORTED error if this bit is set.<37>
	FILE_OPEN_FOR_BACKUP_INTENT = 0x00004000 #The file is being opened for backup intent. That is, it is being opened or created for the purposes of either a backup or a restore operation. The server can check to ensure that the caller is capable of overriding whatever security checks have been placed on the file to allow a backup or restore operation to occur. The server can check for access rights to the file before checking the DesiredAccess field.
	FILE_NO_COMPRESSION = 0x00008000 #The file cannot be compressed. This bit is ignored when FILE_DIRECTORY_FILE is set in CreateOptions.
	FILE_OPEN_REMOTE_INSTANCE = 0x00000400 #This bit SHOULD be set to 0 and MUST be ignored by the server.
	FILE_OPEN_REQUIRING_OPLOCK = 0x00010000 #This bit SHOULD be set to 0 and MUST be ignored by the server.
	FILE_DISALLOW_EXCLUSIVE = 0x00020000 #This bit SHOULD be set to 0 and MUST be ignored by the server.
	FILE_RESERVE_OPFILTER = 0x00100000 #This bit SHOULD be set to 0 and the server MUST fail the request with a STATUS_NOT_SUPPORTED error if this bit is set.<38>
	FILE_OPEN_REPARSE_POINT = 0x00200000 #If the file or directory being opened is a reparse point, open the reparse point itself rather than the target that the reparse point references.
	FILE_OPEN_NO_RECALL = 0x00400000 #In an HSM (Hierarchical Storage Management) environment, this flag means the file SHOULD NOT be recalled from tertiary storage such as tape. The recall can take several minutes. The caller can specify this flag to avoid those delays.
	FILE_OPEN_FOR_FREE_SPACE_QUERY = 0x00800000 #Open file to query for free space. The client SHOULD set this to 0 and the server MUST ignore it.<39>


class CreateAction(enum.Enum):
	FILE_SUPERSEDED = 0x00000000 # An existing file was deleted and a new file was created in its place.
	FILE_OPENED = 0x00000001 #An existing file was opened.
	FILE_CREATED = 0x00000002 #A new file was created.
	FILE_OVERWRITTEN = 0x00000003

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/e8fb45c1-a03d-44ca-b7ae-47385cfd7997
class CREATE_REQ:
	def __init__(self):
		self.StructureSize = 57
		self.SecurityFlags  = 0
		self.RequestedOplockLevel  = None
		self.ImpersonationLevel  = None
		self.SmbCreateFlags  = 0
		self.Reserved   = 0
		self.DesiredAccess    = None
		self.FileAttributes     = None
		self.ShareAccess      = None
		self.CreateDisposition       = None
		self.CreateOptions        = None
		self.NameOffset         = None
		self.NameLength          = None
		self.CreateContextsOffset = 0
		self.CreateContextsLength = 0
		
		self.Buffer = None
		
		#high-level
		self.Name = None
		self.CreateContext = None

	def to_bytes(self):
		if self.Name is not None:
			t_name = self.Name.encode('utf-16-le')
			self.Buffer = t_name
			self.NameOffset = 64 + 2 + 1 + 1 + 4 + 8 + 8 + 4 + 4 + 4 + 4 + 4 + 2 + 2 + 4 + 4
			self.NameLength = len(t_name)
			
		if self.CreateContext:
			t_ctx = self.CreateContext.to_bytes()
			self.CreateContextsLength = len(t_ctx)
			
			pos = curself.NameOffset + self.NameLength
			t_m = pos % 8
			if t_m != 0:
				pad = b'\x00' * (8 - t_m)
				self.Buffer += pad
				self.Buffer +=t_ctx
				self.CreateContextsOffset = pos + len(pad)
			else:
				self.Buffer += t_ctx
				self.CreateContextsOffset = pos
			
		
		t  = self.StructureSize.to_bytes(2, byteorder='little', signed = False)
		t += self.SecurityFlags.to_bytes(1, byteorder='little', signed = False)
		t += self.RequestedOplockLevel.value.to_bytes(1, byteorder='little', signed = False)
		t += self.ImpersonationLevel.value.to_bytes(4, byteorder='little', signed = False)
		t += self.SmbCreateFlags.to_bytes(8, byteorder='little', signed = False)
		t += self.Reserved.to_bytes(8, byteorder='little', signed = False)
		t += self.DesiredAccess.to_bytes(4, byteorder='little', signed = False)
		t += self.FileAttributes.to_bytes(4, byteorder='little', signed = False)
		t += self.ShareAccess.to_bytes(4, byteorder='little', signed = False)
		t += self.CreateDisposition.value.to_bytes(4, byteorder='little', signed = False)
		t += self.CreateOptions.to_bytes(4, byteorder='little', signed = False)
		t += self.NameOffset.to_bytes(2, byteorder='little', signed = False)
		t += self.NameLength.to_bytes(2, byteorder='little', signed = False)
		t += self.CreateContextsOffset.to_bytes(4, byteorder='little', signed = False)
		t += self.CreateContextsLength.to_bytes(4, byteorder='little', signed = False)
		t += self.Buffer
		t += b'\x00'
		return t

	@staticmethod
	def from_bytes(bbuff):
		return CREATE_REQ.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = CREATE_REQ()
		msg.StructureSize = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 57
		msg.SecurityFlags = int.from_bytes(buff.read(1), byteorder='little')
		msg.RequestedOplockLevel = OplockLevel(int.from_bytes(buff.read(1), byteorder='little'))
		msg.ImpersonationLevel = ImpersonationLevel(int.from_bytes(buff.read(4), byteorder = 'little'))
		msg.SmbCreateFlags = int.from_bytes(buff.read(8), byteorder = 'little')
		msg.Reserved = int.from_bytes(buff.read(8), byteorder = 'little')
		msg.DesiredAccess = FileAccessMask(int.from_bytes(buff.read(4), byteorder = 'little'))
		msg.FileAttributes = FileAttributes(int.from_bytes(buff.read(4), byteorder = 'little'))
		msg.ShareAccess = ShareAccess(int.from_bytes(buff.read(4), byteorder = 'little'))
		msg.CreateDisposition = CreateDisposition(int.from_bytes(buff.read(4), byteorder = 'little'))
		msg.CreateOptions  = CreateOptions(int.from_bytes(buff.read(4), byteorder = 'little'))
		msg.NameOffset = int.from_bytes(buff.read(2), byteorder = 'little') #first 8-byte aligned  !!!!!!
		msg.NameLength = int.from_bytes(buff.read(2), byteorder = 'little')
		msg.CreateContextsOffset = int.from_bytes(buff.read(4), byteorder = 'little') #first 8-byte aligned  !!!!!!
		msg.CreateContextsLength = int.from_bytes(buff.read(4), byteorder = 'little')
		
		if msg.NameLength > 0:
			buff.seek(msg.NameOffset, io.SEEK_SET)
			t = buff.read(msg.NameLength)
			msg.Name = t.decode('utf-16-le')
		
		if msg.CreateContextsLength > 0:
			buff.seek(msg.CreateContextsOffset, io.SEEK_SET)
			t = buff.read(msg.CreateContextsLength)
			msg.CreateContext = t #ADDITIONAL PARSING NEEDED BUT NOT YET IMPLEMENTED!!!! See: https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/75364667-3a93-4e2c-b771-592d8d5e876d
		
		return msg

	def __repr__(self):
		t = '==== SMB2 CREATE REQ ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'RequestedOplockLevel: %s\r\n' % self.RequestedOplockLevel
		t += 'ImpersonationLevel: %s\r\n' % self.ImpersonationLevel
		t += 'DesiredAccess: %s\r\n' % self.DesiredAccess
		t += 'FileAttributes: %s\r\n' % self.FileAttributes
		t += 'ShareAccess: %s\r\n' % self.ShareAccess
		t += 'CreateDisposition: %s\r\n' % self.CreateDisposition
		t += 'CreateOptions: %s\r\n' % self.CreateOptions
		t += 'Name: %s\r\n' % self.Name
		t += 'CreateContext: %s\r\n' % self.CreateContext
		return t
		
		
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/d166aa9e-0b53-410e-b35e-3933d8131927
class CREATE_REPLY:
	def __init__(self):
		self.StructureSize = 89
		self.OplockLevel = None
		self.Flags = 0
		self.CreateAction = None
		self.CreationTime   = 0
		self.LastAccessTime    = 0
		self.LastWriteTime     = None
		self.ChangeTime      = None
		self.AllocationSize       = None
		self.EndofFile        = None
		self.FileAttributes         = None
		self.Reserved2          = 0
		self.FileId           = None
		self.CreateContextsOffset = None
		self.CreateContextsLength = None
		
		self.Buffer = b''
		
		#high-level
		self.CreateContext = None

	def to_bytes(self):
		# TODO	
		#return t
		pass

	@staticmethod
	def from_bytes(bbuff):
		return CREATE_REPLY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = CREATE_REPLY()
		msg.StructureSize = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 89
		msg.OplockLevel  = OplockLevel(int.from_bytes(buff.read(1), byteorder='little'))
		msg.Flags  = CreateOptions(int.from_bytes(buff.read(1), byteorder='little'))
		msg.CreateAction  = CreateAction(int.from_bytes(buff.read(4), byteorder = 'little'))
		
		msg.CreationTime  = int.from_bytes(buff.read(8), byteorder = 'little')
		msg.LastAccessTime  = int.from_bytes(buff.read(8), byteorder = 'little')
		msg.LastWriteTime  = int.from_bytes(buff.read(8), byteorder = 'little')
		msg.ChangeTime  = int.from_bytes(buff.read(8), byteorder = 'little')
		msg.AllocationSize  = int.from_bytes(buff.read(8), byteorder = 'little')
		msg.EndofFile  = int.from_bytes(buff.read(8), byteorder = 'little')
		msg.FileAttributes   = FileAttributes(int.from_bytes(buff.read(4), byteorder = 'little'))
		msg.Reserved2  = int.from_bytes(buff.read(4), byteorder = 'little') #first 8-byte aligned  !!!!!!
		msg.FileId  = int.from_bytes(buff.read(16), byteorder = 'little')
		msg.CreateContextsOffset = int.from_bytes(buff.read(4), byteorder = 'little') #first 8-byte aligned  !!!!!!
		msg.CreateContextsLength = int.from_bytes(buff.read(4), byteorder = 'little')
		
		if msg.CreateContextsLength > 0:
			buff.seek(msg.CreateContextsOffset, io.SEEK_SET)
			t = buff.read(msg.CreateContextsLength)
			msg.CreateContext = t #ADDITIONAL PARSING NEEDED BUT NOT YET IMPLEMENTED!!!! See: https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/75364667-3a93-4e2c-b771-592d8d5e876d
		
		return msg

	def __repr__(self):
		t = '==== SMB2 CREATE REPLY ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'OplockLevel: %s\r\n' % self.OplockLevel
		t += 'Flags: %s\r\n' % self.Flags
		t += 'CreationTime: %s\r\n' % self.CreationTime
		t += 'LastAccessTime: %s\r\n' % self.LastAccessTime
		t += 'LastWriteTime: %s\r\n' % self.LastWriteTime
		t += 'ChangeTime: %s\r\n' % self.ChangeTime
		t += 'AllocationSize: %s\r\n' % self.AllocationSize
		t += 'EndofFile: %s\r\n' % self.EndofFile
		t += 'FileAttributes: %s\r\n' % self.FileAttributes
		t += 'Reserved2: %s\r\n' % self.Reserved2
		t += 'FileId: %s\r\n' % self.FileId
		t += 'CreateContextsOffset: %s\r\n' % self.CreateContextsOffset
		t += 'CreateContextsLength: %s\r\n' % self.CreateContextsLength
		return t


import enum

class ACCESS_MASK(enum.IntFlag):
	GENERIC_READ = 0x80000000
	GENERIC_WRITE = 0x4000000
	GENERIC_EXECUTE = 0x20000000
	GENERIC_ALL = 0x10000000
	MAXIMUM_ALLOWED = 0x02000000
	ACCESS_SYSTEM_SECURITY = 0x01000000
	SYNCHRONIZE = 0x00100000
	WRITE_OWNER = 0x00080000
	WRITE_DACL = 0x00040000
	READ_CONTROL = 0x00020000
	DELETE = 0x00010000



from aiosmb.protocol.smb2.headers.transform import *
from aiosmb.protocol.smb2.headers.compression import *
from aiosmb.protocol.smb2.headers.asynch import *
from aiosmb.protocol.smb2.headers.sync import *
from aiosmb.protocol.smb2.headers.common import *


__all__ = ['SMB2HeaderFlag','SMB2Header_SYNC','SMB2Header_ASYNC','SMB2Header_COMPRESSION_TRANSFORM','SMB2Header_TRANSFORM']


import enum
import asyncio

from aiosmb import logger
from aiosmb.commons.exceptions import *

class TCPSocket:
	"""
	Generic asynchronous TCP socket class, nothing SMB related.
	Creates the connection and channels incoming/outgoing bytes via asynchonous queues.
	"""
	def __init__(self, socket = None, target = None):
		self.settings = target
		self.socket = socket #for future, if we want a custom soscket
		self.reader = None
		self.writer = None
		
		self.out_queue = asyncio.Queue()
		self.in_queue = asyncio.Queue()
		
		self.disconnected = asyncio.Event()

		self.incoming_task = None
		self.outgoing_task = None
		
	async def disconnect(self):
		"""
		Disconnects from the socket.
		Stops the reader and writer streams.
		"""
		if self.outgoing_task is not None:
			self.outgoing_task.cancel()
		if self.incoming_task is not None:
			self.incoming_task.cancel()

		if self.writer is not None:
			try:
				self.writer.close()
			except:
				pass
		self.writer = None
		self.reader = None
		self.disconnected.set()

	async def handle_incoming(self):
		"""
		Reads data bytes from the socket and dispatches it to the incoming queue
		"""
		while not self.disconnected.is_set():			
			data = await asyncio.gather(*[self.reader.read(4096)], return_exceptions = True)
			if isinstance(data[0], bytes):
				#print('%s : %s' % (self.writer.get_extra_info('peername')[0], data[0]))
				await self.in_queue.put( (data[0], None) )
			
			elif isinstance(data[0], asyncio.CancelledError):
				return
				
			elif isinstance(data[0], Exception):
				logger.exception('[TCPSocket] handle_incoming %s' % str(data[0]))
				await self.in_queue.put( (None, data[0]) )
				await self.disconnect()
				return
		
	async def handle_outgoing(self):
		"""
		Reads data bytes from the outgoing queue and dispatches it to the socket
		"""
		try:
			while not self.disconnected.is_set():
				data = await self.out_queue.get()
				self.writer.write(data)
				await self.writer.drain()
		except asyncio.CancelledError:
			#the SMB connection is terminating
			return
			
		except Exception as e:
			logger.exception('[TCPSocket] handle_outgoing %s' % str(e))
			await self.disconnect()
			
		
	#async def connect(self, settings): #not needed parameter because we have networkselector now...
	async def connect(self):
		"""
		Main function to be called, connects to the target specified in settings, and starts reading/writing.
		"""

		#self.settings = settings
		
		con = asyncio.open_connection(self.settings.get_ip(), self.settings.get_port())
		
		try:
			self.reader, self.writer = await asyncio.wait_for(con, int(self.settings.timeout))
		except asyncio.TimeoutError:
			logger.debug('[TCPSocket] Connection timeout')
			raise SMBConnectionTimeoutException()
			
		except ConnectionRefusedError:
			logger.debug('[TCPSocket] Connection refused')
			raise SMBConnectionRefusedException()
			
		except asyncio.CancelledError:
			#the SMB connection is terminating
			raise asyncio.CancelledError
			
		except Exception as e:
			logger.exception('[TCPSocket] connect generic exception')
			raise e
		
		self.incoming_task = asyncio.create_task(self.handle_incoming())
		self.outgoing_task = asyncio.create_task(self.handle_outgoing())


		return

			
			


import io

from aiosmb.authentication.ntlm.structures.fields import Fields
from aiosmb.authentication.ntlm.structures.negotiate_flags import NegotiateFlags
from aiosmb.authentication.ntlm.structures.version import Version
from aiosmb.authentication.ntlm.structures.challenge_response import *

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/033d32cc-88f9-4483-9bf2-b273055038ce
class NTLMAuthenticate:
	def __init__(self, _use_NTLMv2 = True):
		self.Signature = b'NTLMSSP\x00'
		self.MessageType = 3
		self.LmChallengeResponseFields = None
		self.NtChallengeResponseFields = None
		self.DomainNameFields = None
		self.UserNameFields = None
		self.WorkstationFields = None
		self.EncryptedRandomSessionKeyFields = None
		self.NegotiateFlags = None
		self.Version = None
		self.MIC = None
		self.Payload = None

		# high level
		self.LMChallenge = None
		self.NTChallenge = None
		self.DomainName = None
		self.UserName = None
		self.Workstation = None
		self.EncryptedRandomSession = None

		# this is a global variable that needs to be indicated
		self._use_NTLMv2 = _use_NTLMv2
		
	@staticmethod
	def construct(flags, domainname= None, workstationname= None, username= None, encrypted_session= None, lm_response= None, nt_response= None, version = None, mic = b'\x00'*16):
		auth = NTLMAuthenticate()
		auth.Payload = b''
		
		payload_pos = 8+4+8+8+8+8+8+8+4
		if flags & NegotiateFlags.NEGOTIATE_VERSION:
			if not version:
				raise Exception('NEGOTIATE_VERSION set but no Version supplied!')
				
			auth.Version = version
			
			payload_pos += 8
			
		if mic is not None:
			self.MIC = mic
			payload_pos += 16
			
		if lm_response:
			data =  lm_response.to_bytes()
			auth.Payload += data
			auth.LmChallengeResponseFields  = Fields(len(data), payload_pos)
			payload_pos += len(data)
			auth.LMChallenge = lm_response
		else:
			auth.LmChallengeResponseFields  = Fields(0,0)
			
		if nt_response:
			data =  nt_response.to_bytes()
			auth.Payload += data
			auth.NtChallengeResponseFields  = Fields(len(data), payload_pos)
			payload_pos += len(data)
			auth.NTChallenge = nt_response
		else:
			auth.NtChallengeResponseFields  = Fields(0,0)
		
		
		if domainname:
			data =  domainname.encode('utf-16le')
			auth.Payload += data
			auth.DomainNameFields  = Fields(len(data), payload_pos)
			payload_pos += len(data)
			auth.DomainName = domainname
		else:
			auth.DomainNameFields  = Fields(0,0)
		
		if username:
			data =  username.encode('utf-16le')
			auth.Payload += data
			auth.UserNameFields  = Fields(len(data), payload_pos)
			payload_pos += len(data)
			auth.UserName = username
		else:
			auth.UserNameFields  = Fields(0,0)
		
		if workstationname:
			data =  workstationname.encode('utf-16le')
			auth.Payload += data
			auth.WorkstationFields  = Fields(len(data), payload_pos)
			payload_pos += len(data)
			auth.Workstation = workstationname
		else:
			auth.WorkstationFields  = Fields(0,0)
			
		if encrypted_session:
			data =  encrypted_session
			auth.Payload += data
			auth.EncryptedRandomSessionKeyFields  = Fields(len(data), payload_pos)
			payload_pos += len(data)
			auth.EncryptedRandomSession = encrypted_session
		else:
			auth.EncryptedRandomSessionKeyFields  = Fields(0,0)
				
		auth.NegotiateFlags = flags		
		return auth
		
	def to_bytes(self):
		t = b''
		t += self.Signature
		t += self.MessageType.to_bytes(4, byteorder = 'little', signed = False)
		
		t += self.LmChallengeResponseFields.to_bytes()
		t += self.NtChallengeResponseFields.to_bytes()
		t += self.DomainNameFields.to_bytes()
		t += self.UserNameFields.to_bytes()
		t += self.WorkstationFields.to_bytes()
		t += self.EncryptedRandomSessionKeyFields.to_bytes()
		t += self.NegotiateFlags.to_bytes(4, byteorder = 'little', signed = False)
		if self.Version:
			t += self.Version.to_bytes()
		if self.MIC is not None:	
			t += self.MIC
		t += self.Payload
		return t
		

	@staticmethod
	def from_bytes(bbuff,_use_NTLMv2 = True):
		return NTLMAuthenticate.from_buffer(io.BytesIO(bbuff), _use_NTLMv2 = _use_NTLMv2)

	@staticmethod
	def from_buffer(buff, _use_NTLMv2 = True):
		auth = NTLMAuthenticate(_use_NTLMv2)
		auth.Signature    = buff.read(8)
		auth.MessageType  = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		auth.LmChallengeResponseFields = Fields.from_buffer(buff)
		auth.NtChallengeResponseFields = Fields.from_buffer(buff)
		auth.DomainNameFields = Fields.from_buffer(buff)
		auth.UserNameFields = Fields.from_buffer(buff)
		auth.WorkstationFields = Fields.from_buffer(buff)
		auth.EncryptedRandomSessionKeyFields = Fields.from_buffer(buff)
		auth.NegotiateFlags = NegotiateFlags(int.from_bytes(buff.read(4), byteorder = 'little', signed = False))
		if auth.NegotiateFlags & NegotiateFlags.NEGOTIATE_VERSION: 
			auth.Version = Version.from_buffer(buff)

		# TODO: I'm not sure about this condition!!! Need to test this!
		if auth.NegotiateFlags & NegotiateFlags.NEGOTIATE_ALWAYS_SIGN:
			auth.MIC = buff.read(16)

		currPos = buff.tell()
		auth.Payload = buff.read()

		if auth._use_NTLMv2 and auth.NtChallengeResponseFields.length > 24:
			buff.seek(auth.LmChallengeResponseFields.offset, io.SEEK_SET)
			auth.LMChallenge = LMv2Response.from_buffer(buff)
			

			buff.seek(auth.NtChallengeResponseFields.offset, io.SEEK_SET)
			auth.NTChallenge = NTLMv2Response.from_buffer(buff)

		else:
			buff.seek(auth.LmChallengeResponseFields.offset, io.SEEK_SET)
			auth.LMChallenge = LMResponse.from_buffer(buff)
				
			buff.seek(auth.NtChallengeResponseFields.offset, io.SEEK_SET)
			auth.NTChallenge = NTLMv1Response.from_buffer(buff)

		buff.seek(auth.DomainNameFields.offset,io.SEEK_SET)
		auth.DomainName = buff.read(auth.DomainNameFields.length).decode('utf-16le')
		
		buff.seek(auth.UserNameFields.offset,io.SEEK_SET)
		auth.UserName = buff.read(auth.UserNameFields.length).decode('utf-16le')

		buff.seek(auth.WorkstationFields.offset,io.SEEK_SET)
		auth.Workstation = buff.read(auth.WorkstationFields.length).decode('utf-16le')

		buff.seek(auth.EncryptedRandomSessionKeyFields.offset,io.SEEK_SET)
		auth.EncryptedRandomSession = buff.read(auth.EncryptedRandomSessionKeyFields.length)
		
		buff.seek(currPos, io.SEEK_SET)

		return auth

	def __repr__(self):
		t  = '== NTLMAuthenticate ==\r\n'
		t += 'Signature     : %s\r\n' % repr(self.Signature)
		t += 'MessageType   : %s\r\n' % repr(self.MessageType)
		t += 'NegotiateFlags: %s\r\n' % repr(self.NegotiateFlags)
		t += 'Version       : %s\r\n' % repr(self.Version)
		t += 'MIC           : %s\r\n' % repr(self.MIC.hex() if self.MIC else 'None')
		t += 'LMChallenge   : %s\r\n' % repr(self.LMChallenge)
		t += 'NTChallenge   : %s\r\n' % repr(self.NTChallenge)
		t += 'DomainName    : %s\r\n' % repr(self.DomainName)
		t += 'UserName      : %s\r\n' % repr(self.UserName)
		t += 'Workstation   : %s\r\n' % repr(self.Workstation)
		t += 'EncryptedRandomSession: %s\r\n' % repr(self.EncryptedRandomSession.hex())
		return t

def test():
	test_reconstrut()
	test_construct()
	test_2()
	
def test_2():
	data = bytes.fromhex('4e 54 4c 4d 53 53 50 00 03 00 00 00 18 00 18 006c 00 00 00 54 00 54 00 84 00 00 00 0c 00 0c 0048 00 00 00 08 00 08 00 54 00 00 00 10 00 10 005c 00 00 00 10 00 10 00 d8 00 00 00 35 82 88 e205 01 28 0a 00 00 00 0f 44 00 6f 00 6d 00 61 0069 00 6e 00 55 00 73 00 65 00 72 00 43 00 4f 004d 00 50 00 55 00 54 00 45 00 52 00 86 c3 50 97ac 9c ec 10 25 54 76 4a 57 cc cc 19 aa aa aa aaaa aa aa aa 68 cd 0a b8 51 e5 1c 96 aa bc 92 7beb ef 6a 1c 01 01 00 00 00 00 00 00 00 00 00 0000 00 00 00 aa aa aa aa aa aa aa aa 00 00 00 0002 00 0c 00 44 00 6f 00 6d 00 61 00 69 00 6e 0001 00 0c 00 53 00 65 00 72 00 76 00 65 00 72 0000 00 00 00 00 00 00 00 c5 da d2 54 4f c9 79 9094 ce 1c e9 0b c9 d0 3e')
	challenge = NTLMAuthenticate.from_bytes(data)
	print(repr(challenge))
	
def test_reconstrut(data = None):
	print('=== reconstruct===')
	if not data:
		auth_test_data = bytes.fromhex('4e544c4d5353500003000000180018007c000000180118019400000008000800580000000c000c0060000000100010006c00000010001000ac010000158288e20a00d73a0000000f0d98eb57e9c52820709c99b98ca321a15400450053005400760069006300740069006d00570049004e0031003000580036003400000000000000000000000000000000000000000000000000fade3940b9381c53c91ddcdd0d44000b0101000000000000aec600bfc5fdd4011bfa20699d7628730000000002000800540045005300540001001200570049004e003200300031003900410044000400120074006500730074002e0063006f007200700003002600570049004e003200300031003900410044002e0074006500730074002e0063006f007200700007000800aec600bfc5fdd40106000400020000000800300030000000000000000000000000200000527d27f234de743760966384d36f61ae2aa4fc2a380699f8caa600011b486d890a0010000000000000000000000000000000000009001e0063006900660073002f00310030002e00310030002e00310030002e003200000000000000000000000000fd67edfb41c09465a91fd733deb0b55b')
	else:
		auth_test_data = data
	challenge = NTLMAuthenticate.from_bytes(auth_test_data)
	print(repr(challenge))
	auth_test_data_verify = challenge.to_bytes()
	print('====== reconstructed ====')
	print(hexdump(auth_test_data_verify))
	print('====== original ====')
	print(hexdump(auth_test_data))
	assert auth_test_data == auth_test_data_verify
	
	
def test_construct():
	pass
	
if __name__ == '__main__':
	from aiosmb.utils.hexdump import hexdump
	
	test()


import enum
import asyncio

from aiosmb import logger
from aiosmb.exceptions import *
from aiosmb.network.netbios import NetBIOSTransport
from aiosmb.smbconnection_server import SMBServerConnection

class TCPClient:
	def __init__(self, raddr, rport, reader, writer):
		self.raddr = raddr
		self.rport = rport
		self.reader = reader
		self.writer = writer
		self.out_queue = asyncio.Queue()
		self.in_queue = asyncio.Queue()

class TCPServerSocket:
	"""
	Generic asynchronous TCP socket class, nothing SMB related.
	Creates the connection and channels incoming/outgoing bytes via asynchonous queues.
	"""
	def __init__(self, ip, server_settings, port = 445,  shutdown_evt = asyncio.Event(), socket = None):
		self.ip = ip
		self.port = port
		self.server_settings = server_settings
		self.socket = socket #for future, if we want a custom soscket
		
		self.disconnected = asyncio.Event()
		self.shutdown_evt = shutdown_evt
		
	async def handle_incoming(self, client):
		"""
		Reads data bytes from the socket and dispatches it to the incoming queue
		"""
		while not self.shutdown_evt.is_set():			
			data = await asyncio.gather(*[client.reader.read(4096)], return_exceptions = True)
			
			if isinstance(data[0], bytes):
				await client.in_queue.put(data[0])
			
			elif isinstance(data[0], asyncio.CancelledError):
				return
				
			elif isinstance(data[0], Exception):
				logger.exception('[TCPSocket] handle_incoming %s' % str(data[0]))
				return
		
	async def handle_outgoing(self, client):
		"""
		Reads data bytes from the outgoing queue and dispatches it to the socket
		"""
		try:
			while not self.shutdown_evt.is_set():
				data = await client.out_queue.get()
				client.writer.write(data)
				await client.writer.drain()
		except asyncio.CancelledError:
			#the SMB connection is terminating
			return
			
		except Exception as e:
			logger.exception('[TCPSocket] handle_outgoing %s' % str(e))
		
	async def handle_client(self, reader, writer):
		raddr, rport = writer.get_extra_info('peername')
		logger.info('TCP client connected from %s:%s' % (raddr, rport))
		
		client = TCPClient(raddr, rport, reader, writer)
		asyncio.ensure_future(self.handle_outgoing(client))
		asyncio.ensure_future(self.handle_incoming(client))
		nbtransport = NetBIOSTransport(client)
		server = SMBServerConnection(self.server_settings, nbtransport)
		
		await nbtransport.run()
		await server.main()
		logger.info('SMB server terminated, closing client! %s:%s' % (raddr, rport))
		
		
	async def listen(self):
		server = await asyncio.start_server(self.handle_client, self.ip, self.port)
		addr = server.sockets[0].getsockname()
		async with server:
			logger.info('TCP Server in listening state')
			await server.serve_forever()
		logger.info('TCP server terminated')	
			
	async def run(self):
		return self.listen()
		


import enum
import asyncio

from aiosmb import logger
from aiosmb.dcerpc.v5.common.service import SMBServiceStatus, SMBService
from aiosmb.dcerpc.v5.common.connection.smbdcefactory import SMBDCEFactory
from aiosmb.dcerpc.v5 import wkst, scmr

from aiosmb.commons.utils.decorators import red, rr, red_gen

class SMBRemoteServieManager:
	def __init__(self, connection):
		self.connection = connection
		self.dce = None
		self.handle = None
		
		self.service_handles = {} #service_name -> handle
		
	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True,None
	
	@red
	async def connect(self, open = True):
		rpctransport = SMBDCEFactory(self.connection, filename=r'\svcctl')
		self.dce = rpctransport.get_dce_rpc()
		await rr(self.dce.connect())
		await rr(self.dce.bind(scmr.MSRPC_UUID_SCMR))

		if open == True:
			await rr(self.open())
		
		return True,None
	
	@red
	async def open(self):
		if not self.dce:
			await rr(self.connect())
		
		ans, _ = await rr(scmr.hROpenSCManagerW(self.dce))
		self.handle = ans['lpScHandle']

		return True,None
	
	@red
	async def close(self):
		if self.dce:
			if self.handle:
				for service_name in self.service_handles:
					try:
						await self.close_service(service_name)
					except:
						pass
				try:
					await scmr.hRCloseServiceHandle(self.dce, self.service_handles[service_name])
				except:
					pass
			try:
				await self.dce.disconnect()
			except:
				pass
			return
		
		return True,None

	@red_gen
	async def list(self):
		resp, _ = await rr(scmr.hREnumServicesStatusW(self.dce, self.handle))
		for i in range(len(resp)):
			service_status = None
			state = resp[i]['ServiceStatus']['dwCurrentState']
			if state == scmr.SERVICE_CONTINUE_PENDING:
			   service_status = SMBServiceStatus.CONTINUE_PENDING
			elif state == scmr.SERVICE_PAUSE_PENDING:
			   service_status = SMBServiceStatus.PAUSE_PENDING
			elif state == scmr.SERVICE_PAUSED:
			   service_status = SMBServiceStatus.PAUSED
			elif state == scmr.SERVICE_RUNNING:
			   service_status = SMBServiceStatus.RUNNING
			elif state == scmr.SERVICE_START_PENDING:
			   service_status = SMBServiceStatus.START_PENDING
			elif state == scmr.SERVICE_STOP_PENDING:
			   service_status = SMBServiceStatus.STOP_PENDING
			elif state == scmr.SERVICE_STOPPED:
			   service_status = SMBServiceStatus.STOPPED
			else:
			   service_status = SMBServiceStatus.UNKNOWN

			service = SMBService(resp[i]['lpServiceName'][:-1], resp[i]['lpDisplayName'][:-1], service_status)
			yield service, None
	
	@red
	async def open_service(self, service_name):
		if service_name in self.service_handles:
			return False, None
			
		ans, _ = await rr(scmr.hROpenServiceW(self.dce, self.handle, service_name))
		self.service_handles[service_name] = ans['lpServiceHandle']

		return True,None
	
	@red
	async def close_service(self, service_name):
		if not self.handle:
			await rr(self.open())
		if service_name not in self.service_handles:
			await rr(self.open_service(service_name))
		
		await rr(scmr.hRCloseServiceHandle(self.dce, self.service_handles[service_name]))
		del self.service_handles[service_name]

		return True,None
	
	@red
	async def check_service_status(self, service_name):
		if not self.handle:
			await rr(self.open())
		if service_name not in self.service_handles:
			await rr(self.open_service(service_name))
		
		# Let's check its status
		ans, _ = await rr(scmr.hRQueryServiceStatus(self.dce, self.service_handles[service_name]))
		if ans['lpServiceStatus']['dwCurrentState'] == scmr.SERVICE_STOPPED:
			logger.info('Service %s is in stopped state'% service_name)
			
			# Let's check its configuration if service is stopped, maybe it's disabled :s
			ans, _ = await rr(scmr.hRQueryServiceConfigW(self.dce,self.handle))
			if ans['lpServiceConfig']['dwStartType'] == 0x4:
				logger.info('Service %s is disabled'% service_name)
				return SMBServiceStatus.DISABLED, None
			else:
				return SMBServiceStatus.STOPPED, None

		elif ans['lpServiceStatus']['dwCurrentState'] == scmr.SERVICE_RUNNING:
			logger.debug('Service %s is already running'% service_name)
			return SMBServiceStatus.RUNNING, None
		else:
			raise Exception('Unknown service state 0x%x - Aborting' % ans['CurrentState'])

		return False, None
	
	@red
	async def stop_service(self, service_name):
		raise NotImplementedError('stop_service')
	
	@red
	async def create_service(self, service_name, display_name, command):
		if not self.handle:
			await rr(self.open())
		#print(service_name)
		#print(display_name)
		#print(command)
		resp, _ = await rr(scmr.hRCreateServiceW(self.dce, self.handle, service_name + '\x00', display_name + '\x00', lpBinaryPathName=command + '\x00'))
		self.service_handles[service_name] = resp['lpServiceHandle']
		return True,None
	
	@red
	async def delete_service(self, service_name):
		if not self.handle:
			await rr(self.open())
		if service_name not in self.service_handles:
			await rr(self.open_service(service_name))
		
		await rr(scmr.hRDeleteService(self.dce, self.service_handles[service_name]))
		return True,None
	
	@red
	async def start_service(self, service_name):
		if not self.handle:
			await rr(self.open())
		if service_name not in self.service_handles:
			await rr(self.open_service(service_name))

			await rr(scmr.hRStartServiceW(self.dce , self.service_handles[service_name]))
			await asyncio.sleep(1) #service takes time to start up...
		
		return True,None
	
	@red
	async def enable_service(self, service_name):
		if not self.handle:
			await rr(self.open())
		if service_name not in self.service_handles:
			await rr(self.open_service(service_name))
			
		await rr(scmr.hRChangeServiceConfigW(self.dce, self.service_handles[service_name]))
		return True,None



import os
import io

# https://docs.microsoft.com/en-us/previous-versions/aa373931(v%3Dvs.80)
# Dear Microsoft ppl, pls fuck youselves...
class GUID:
	def __init__(self):
		self.Data1 = None
		self.Data2 = None
		self.Data3 = None
		self.Data4 = None
		
	@staticmethod
	def from_bytes(data):
		return GUID.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		guid = GUID()
		guid.Data1 = buff.read(4)[::-1]
		guid.Data2 = buff.read(2)[::-1]
		guid.Data3 = buff.read(2)[::-1]
		guid.Data4 = buff.read(8)
		return guid
		
	@staticmethod
	def from_string(str):
		guid = GUID()
		guid.Data1 = bytes.fromhex(str.split('-')[0])
		guid.Data2 = bytes.fromhex(str.split('-')[1])
		guid.Data3 = bytes.fromhex(str.split('-')[2])
		guid.Data4 = bytes.fromhex(str.split('-')[3])
		guid.Data4 += bytes.fromhex(str.split('-')[4])
		return guid	
		
	def to_bytes(self):
		t =  self.Data1
		t +=  self.Data2
		t +=  self.Data3
		t +=  self.Data4	
		return t

	@staticmethod
	def random():
		guid = GUID()
		guid.Data1 = os.urandom(4)
		guid.Data2 = os.urandom(2)
		guid.Data3 = os.urandom(2)
		guid.Data4 = os.urandom(8)
		return guid
		
	def __str__(self):
		return '-'.join([self.Data1.hex(), self.Data2.hex(),self.Data3.hex(),self.Data4[:2].hex(),self.Data4[2:].hex()])


import datetime
import io

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/2c57429b-fdd4-488f-b5fc-9e4cf020fcdf
class FILETIME:
	def __init__(self):
		self.dwLowDateTime = None
		self.dwHighDateTime = None
		
		self.datetime = None
	@staticmethod
	def from_bytes(data):
		return FILETIME.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		t = FILETIME()
		t.dwLowDateTime = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		t.dwHighDateTime = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		if t.dwHighDateTime == 4294967295 and t.dwLowDateTime == 4294967295:
			t.datetime = t.datetime = datetime.datetime(3000, 1, 1, 0, 0)
		else:
			ft = (t.dwHighDateTime << 32) + t.dwLowDateTime
			if ft == 0:
				t.datetime = datetime.datetime(1970, 1, 1, 0, 0)
			else:
				t.datetime = datetime.datetime.utcfromtimestamp((ft - 116444736000000000) / 10000000)
		return t


import asyncio

from aiosmb import logger
from aiosmb.commons.exceptions import *
from aiosmb.commons.connection.proxy import SMBProxyType
from aiosmb.network.socks import SocksProxyConnection
from aiosmb.network.multiplexornetwork import MultiplexorProxyConnection
from aiosmb.commons.utils.decorators import red, rr

class DCERPCTCPConnection:
	def __init__(self, ip, port):
		self.ip = ip
		self.port = port
		self.timeout = 2
		
		self.in_queue = asyncio.Queue()
		self.out_queue = asyncio.Queue()
		
		
		self.disconnected = asyncio.Event()
		self.shutdown_evt = asyncio.Event()

		self.__incoming_task = None
		self.__outgoing_task = None
		
	
	async def handle_incoming(self):
		"""
		Reads data bytes from the socket and dispatches it to the incoming queue
		"""
		try:
			while not self.disconnected.is_set() or not self.shutdown_evt.is_set():
				data = await self.reader.read(4096)
				await self.in_queue.put((data, None))
		
		except asyncio.CancelledError:
			#the SMB connection is terminating
			return
			
		except Exception as e:	
			logger.exception('[DCERPCTCPConnection] handle_incoming %s' % str(e))
			await self.in_queue.put((None, e))
			await self.disconnect()
			
		
	async def handle_outgoing(self):
		"""
		Reads data bytes from the outgoing queue and dispatches it to the socket
		"""
		try:
			while not self.disconnected.is_set() or not self.shutdown_evt.is_set():
				data = await self.out_queue.get()
				self.writer.write(data)
				await self.writer.drain()
		except asyncio.CancelledError:
			#the connection is terminating
			return
			
		except Exception as e:
			logger.exception('[DCERPCTCPConnection] handle_outgoing %s' % str(e))
			await self.disconnect()
	
	@red
	async def connect(self):
		con = asyncio.open_connection(self.ip, self.port)
		
		try:
			self.reader, self.writer = await asyncio.wait_for(con, int(self.timeout))
		except asyncio.TimeoutError:
			logger.debug('[DCERPCTCPConnection] Connection timeout')
			raise SMBConnectionTimeoutException() 
		except ConnectionRefusedError:
			logger.debug('[DCERPCTCPConnection] Connection refused')
			raise SMBConnectionRefusedException()
		except asyncio.CancelledError:
			#the SMB connection is terminating
			return
		except Exception as e:
			logger.exception('[DCERPCTCPConnection] connect generic exception')
			raise e
		

		self.__incoming_task = asyncio.create_task(self.handle_incoming())
		self.__outgoing_task = asyncio.create_task(self.handle_outgoing())
		
		
		return True, None
	
	@red
	async def disconnect(self):
		"""
		Disconnects from the socket.
		Stops the reader and writer streams.
		"""
		if self.__incoming_task:
			self.__incoming_task.cancel()
		if self.__outgoing_task:
			self.__outgoing_task.cancel()

		self.reader = None
		try:
			self.writer.close()
		except:
			pass
		self.writer = None
		self.disconnected.set()

		return True, None


class DCERPCTCPTransport:
	def __init__(self, target):
		self.target = target
		
		self.buffer = b''
		self.data_in_evt = asyncio.Event()
		self.exception_evt = asyncio.Event()
		self.connection = None

		self.__last_exception = None
		self.__incoming_task = None
		self.__outgoing_task = None

		self._max_send_frag = 1024

	@red
	async def get_connection_layer(self):
		if self.target.proxy is None:
			return DCERPCTCPConnection(self.target.ip, self.target.port), None
			
		elif self.target.proxy.type in [SMBProxyType.SOCKS5, SMBProxyType.SOCKS5_SSL, SMBProxyType.SOCKS4, SMBProxyType.SOCKS4_SSL]:
			return SocksProxyConnection(target = self.target), None

		elif self.target.proxy.type in [SMBProxyType.MULTIPLEXOR, SMBProxyType.MULTIPLEXOR_SSL]:
			mpc = MultiplexorProxyConnection(self.target)
			socks_proxy = await mpc.connect()
			return socks_proxy, None

		else:
			raise Exception('Unknown proxy type %s' % self.target.proxy.type)


	async def __handle_incoming(self):
		try:
			while True:
				data, res = await self.connection.in_queue.get()
				if data is None:
					self.__last_exception = res
					self.exception_evt.set()
					self.data_in_evt.set()
					return
				self.buffer += data
				self.data_in_evt.set()
		except asyncio.CancelledError:
			self.exception_evt.set()
			return
		except Exception as e:
			logger.exception('__handle_incoming')
			self.exception_evt.set()
			return
	
	@red
	async def connect(self):
		self.connection, _ = await rr(self.get_connection_layer())
		await self.connection.connect()

		self.__incoming_task = asyncio.create_task(self.__handle_incoming())
		
		return True, None
	
	@red
	async def disconnect(self):
		try:
			if self.__incoming_task:
				self.__incoming_task.cancel()
			if self.data_in_evt:
				self.data_in_evt.set()
			await self.connection.disconnect()
		except:
			pass

		return True, None
	
	@red
	async def send(self, data, forceWriteAndx = 0, forceRecv = 0):
		if self.__last_exception is not None:
			return None, self.__last_exception
		
		if self._max_send_frag:
			offset = 0
			while True:
				toSend = data[offset:offset+self._max_send_frag]
				if not toSend:
					break
				await self.connection.out_queue.put(toSend)
				offset += len(toSend)
				
		else:
			await self.connection.out_queue.put(data)
		
		return True, None
	
	@red
	async def recv(self, count, forceRecv = 0):
		try:
			while len(self.buffer) < count:
				#waiting until buffer has enough data
				self.data_in_evt.clear()
				await asyncio.wait_for(self.data_in_evt.wait(), timeout = self.target.timeout)
			
			if self.__last_exception is not None:
				return None, self.__last_exception

			data = self.buffer[:count]
			self.buffer = self.buffer[count:]
			return data, None
		except Exception as e:
			return None, e
		


import io
import enum

from aiosmb.protocol.smb.command_codes import SMBCommand
from aiosmb.wintypes.ntstatus import NTStatus


class SMBHeaderFlagsEnum(enum.IntFlag):
	SMB_FLAGS_LOCK_AND_READ_OK = 0x01
	SMB_FLAGS_BUF_AVAIL = 0x02
	Reserved = 0x04
	SMB_FLAGS_CASE_INSENSITIVE = 0x08
	SMB_FLAGS_CANONICALIZED_PATHS = 0x10
	SMB_FLAGS_OPLOCK = 0x20
	SMB_FLAGS_OPBATCH = 0x40
	SMB_FLAGS_REPLY = 0x80


class SMBHeaderFlags2Enum(enum.IntFlag):
	SMB_FLAGS2_LONG_NAMES = 0x0001
	SMB_FLAGS2_EAS = 0x0002
	SMB_FLAGS2_SMB_SECURITY_SIGNATURE = 0x0004
	SMB_FLAGS2_IS_LONG_NAME = 0x0040
	SMB_FLAGS2_EXTENDED_SECURITY = 0x0800
	SMB_FLAGS2_DFS = 0x1000
	SMB_FLAGS2_PAGING_IO = 0x2000
	SMB_FLAGS2_NT_STATUS = 0x4000
	SMB_FLAGS2_UNICODE = 0x8000

class SMBHeader:
	def __init__(self):
		self.Protocol = b'\xFFSMB'
		self.Command  = None
		self.Status   = None
		self.Flags    = None
		self.Flags2   = None
		self.PIDHigh  = 0
		self.SecurityFeatures = b'\x00' * 8
		self.Signature = b'\x00' * 8
		self.Reserved = 0
		self.TID      = 65535
		self.PIDLow   = 0
		self.UID      = 0
		self.MessageId      = 0

	@staticmethod
	def from_bytes(bbuff):
		return SMBHeader.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		hdr = SMBHeader()
		hdr.Protocol = buff.read(4)
		assert hdr.Protocol == b'\xFFSMB', "SMBv1 Header Magic incorrect!"
		hdr.Command  = SMBCommand(int.from_bytes(buff.read(1), byteorder='little', signed = False))
		hdr.Status   = NTStatus(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		hdr.Flags    = SMBHeaderFlagsEnum(int.from_bytes(buff.read(1), byteorder='little', signed = False))
		hdr.Flags2   = SMBHeaderFlags2Enum(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		hdr.PIDHigh  = int.from_bytes(buff.read(2), byteorder='little', signed = False)

		if SMBHeaderFlags2Enum.SMB_FLAGS2_SMB_SECURITY_SIGNATURE in hdr.Flags2:
			hdr.SecurityFeatures = buff.read(8)
		else:
			hdr.Signature = buff.read(8)

		hdr.Reserved = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		hdr.TID      = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		hdr.PIDLow   = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		hdr.UID      = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		hdr.MessageId      = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		return hdr

	@staticmethod
	def construct(command, status, flags, flags2, uid = 0, mid = 0, tid = 0, securityfeatures = None, signature = None, pidhigh = 0, pidlow = 0):
		hdr = SMBHeader()
		hdr.Protocol = b'\xFFSMB'
		hdr.Command  = command
		hdr.Status   = status
		hdr.Flags    = flags
		hdr.Flags2   = flags2
		hdr.PIDHigh  = pidhigh

		if SMBHeaderFlags2Enum.SMB_FLAGS2_SMB_SECURITY_SIGNATURE in hdr.Flags2:
			if securityfeatures is None:
				raise Exception('SMB_FLAGS2_SMB_SECURITY_SIGNATURE is present but SecurityFeatures was not supplied!')
			hdr.SecurityFeatures = securityfeatures
		else:
			if signature is not None:
				hdr.Signature = signature

		hdr.Reserved = 0
		hdr.TID      = tid
		hdr.PIDLow   = pidlow
		hdr.UID      = uid
		hdr.MessageId      = mid

		return hdr

	def to_bytes(self):
		t  = self.Protocol
		t += self.Command.value.to_bytes(1, byteorder = 'little', signed=False)
		t += self.Status.value.to_bytes(4, byteorder = 'little', signed=False)
		t += self.Flags.to_bytes(1, byteorder = 'little', signed=False)
		t += self.Flags2.value.to_bytes(2, byteorder = 'little', signed=False)
		t += self.PIDHigh.to_bytes(2, byteorder = 'little', signed=False)
		if self.SecurityFeatures is not None:
			t += self.SecurityFeatures
		elif self.Signature is not None:
			t += self.Signature
		else:
			t += b'\x00'*8
		t += self.Reserved.to_bytes(2, byteorder = 'little', signed=False)
		t += self.TID.to_bytes(2, byteorder = 'little', signed=False)
		t += self.PIDLow.to_bytes(2, byteorder = 'little', signed=False)
		t += self.UID.to_bytes(2, byteorder = 'little', signed=False)
		t += self.MessageId.to_bytes(2, byteorder = 'little', signed=False)
		return t

	def __repr__(self):
		t = '===SMBHeader===\r\n'
		t += 'Command: %s\r\n' % self.Command.name
		t += 'Flags:   %s\r\n' % repr(self.Flags)
		t += 'Flags2:  %s\r\n' % repr(self.Flags2)
		t += 'PIDHigh: %s\r\n' % self.PIDHigh
		t += 'SecurityFeatures: %s\r\n' % (self.SecurityFeatures.hex() if self.SecurityFeatures is not None else 'NONE')
		t += 'Reserved: %s\r\n' % self.Reserved
		t += 'TID: %s\r\n' % self.TID
		t += 'PIDLow: %s\r\n' % self.PIDLow
		t += 'UID: %s\r\n' % self.UID
		t += 'MID: %s\r\n' % self.MessageId
		return t


import io
from aiosmb.wintypes.dtyp.structures.filetime import FILETIME
from aiosmb.wintypes.fscc.FileAttributes import FileAttributes

class FileFullDirectoryInformationList:
	def __init__(self):
		pass
		
	@staticmethod
	def from_bytes(data):
		return FileFullDirectoryInformationList.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		pos = buff.tell()
		t = []
		info = FileFullDirectoryInformation.from_buffer(buff)
		t.append(info)
		while info.NextEntryOffset != 0:
			buff.seek(pos + info.NextEntryOffset, 0)
			pos = buff.tell()
			info = FileFullDirectoryInformation.from_buffer(buff)
			t.append(info)
			
		return t
			

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/e8d926d1-3a22-4654-be9c-58317a85540b
class FileFullDirectoryInformation:
	def __init__(self):
		self.NextEntryOffset = None
		self.FileIndex  = None
		self.CreationTime  = None
		self.LastAccessTime  = None
		self.LastWriteTime  = None
		self.ChangeTime  = None
		self.EndOfFile  = None
		self.AllocationSize  = None
		self.FileAttributes   = None
		self.FileNameLength   = None
		self.EaSize   = None
		self.FileName   = None
		
	@staticmethod
	def from_bytes(data):
		return FileFullDirectoryInformation.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		msg = FileFullDirectoryInformation()
		msg.NextEntryOffset = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.FileIndex = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		msg.CreationTime = FILETIME.from_buffer(buff).datetime
		msg.LastAccessTime = FILETIME.from_buffer(buff).datetime
		msg.LastWriteTime = FILETIME.from_buffer(buff).datetime
		msg.ChangeTime = FILETIME.from_buffer(buff).datetime
		msg.EndOfFile = int.from_bytes(buff.read(8), byteorder = 'little', signed = True)
		msg.AllocationSize = int.from_bytes(buff.read(8), byteorder = 'little', signed = True)
		msg.FileAttributes = FileAttributes(int.from_bytes(buff.read(4), byteorder = 'little', signed = False))
		msg.FileNameLength = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		msg.EaSize = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		msg.FileName = buff.read(msg.FileNameLength).decode('utf-16-le')
		return msg
		
	def __str__(self):
		t = '===== %s =====\r\n' % 'FileFullDirectoryInformation'
		d = self.__dict__
		for k in d:
			t += '%s : %s\r\n' % (k, d[k])
		return t


import io
import enum

from aiosmb.wintypes.fscc.structures.fileinfoclass import *
	
class QueryInfoType(enum.Enum):
	FILE = 0x01 #The file information is requested.
	FILESYSTEM = 0x02 #The underlying object store information is requested.
	SECURITY = 0x03 #The security information is requested.
	QUOTA = 0x04
	
class SecurityInfo(enum.IntFlag):
	OWNER_SECURITY_INFORMATION = 0x00000001 #The client is querying the owner from the security descriptor of the file or named pipe.
	GROUP_SECURITY_INFORMATION = 0x00000002 #The client is querying the group from the security descriptor of the file or named pipe.
	DACL_SECURITY_INFORMATION = 0x00000004 #The client is querying the discretionary access control list from the security descriptor of the file or named pipe.
	SACL_SECURITY_INFORMATION = 0x00000008 #The client is querying the system access control list from the security descriptor of the file or named pipe.
	LABEL_SECURITY_INFORMATION = 0x00000010 #The client is querying the integrity label from the security descriptor of the file or named pipe.
	ATTRIBUTE_SECURITY_INFORMATION = 0x00000020 #The client is querying the resource attribute from the security descriptor of the file or named pipe.
	SCOPE_SECURITY_INFORMATION = 0x00000040 #The client is querying the central access policy of the resource from the security descriptor of the file or named pipe.
	BACKUP_SECURITY_INFORMATION = 0x00010000 #The client is querying the security descriptor information used for backup operation.
	
class EaInformation(enum.IntFlag):
	SL_RESTART_SCAN = 0x00000001 #Restart the scan for EAs from the beginning.
	SL_RETURN_SINGLE_ENTRY = 0x00000002 #Return a single EA entry in the response buffer.
	SL_INDEX_SPECIFIED = 0x00000004 #The caller has specified an EA index.

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/d623b2f7-a5cd-4639-8cc9-71fa7d9f9ba9
class QUERY_INFO_REQ:
	def __init__(self):
		self.StructureSize = 41
		self.InfoType = None
		self.FileInfoClass = None
		self.OutputBufferLength = 65535
		self.InputBufferOffset = 0
		self.Reserved = 0
		self.InputBufferLength = 0
		self.AdditionalInformation = None
		self.Flags = None
		self.FileId = None
		self.Buffer = None
		
		#high-level param
		self.Data = None
		
	def to_bytes(self):
		if self.Data:
			self.Buffer = io.BytesIO(self.Data)
			self.InputBufferOffset = 64 + 2+1+1+4+2+2+2+4+4+16
			self.InputBufferLength = len(self.Data)
			
		t  = self.StructureSize.to_bytes(2, byteorder='little', signed = False)
		t += self.InfoType.value.to_bytes(1, byteorder='little', signed = False)
		t += self.FileInfoClass.value.to_bytes(1, byteorder='little', signed = False)
		t += self.OutputBufferLength.to_bytes(4, byteorder='little', signed = False)
		t += self.InputBufferOffset.to_bytes(2, byteorder='little', signed = False)
		t += self.Reserved.to_bytes(2, byteorder='little', signed = False)
		t += self.InputBufferLength.to_bytes(4, byteorder='little', signed = False)
		t += self.AdditionalInformation.to_bytes(4, byteorder='little', signed = False)
		t += self.Flags.to_bytes(4, byteorder='little', signed = False)
		t += self.FileId.to_bytes(16, byteorder='little', signed = False)
		
		if self.Buffer:
			t += self.Buffer
		return t

	@staticmethod
	def from_bytes(bbuff):
		return QUERY_INFO_REQ.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = QUERY_INFO_REQ()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 41
		msg.InfoType = QueryInfoType(int.from_bytes(buff.read(1), byteorder='little'))
		msg.FileInfoClass  = FileInfoClass(int.from_bytes(buff.read(1), byteorder='little'))
		msg.OutputBufferLength  = int.from_bytes(buff.read(4), byteorder = 'little')
		msg.InputBufferOffset  = int.from_bytes(buff.read(2), byteorder = 'little')
		msg.Reserved   = int.from_bytes(buff.read(2), byteorder = 'little')
		msg.InputBufferLength    = int.from_bytes(buff.read(2), byteorder = 'little')
		msg.AdditionalInformation = SecurityInfo(int.from_bytes(buff.read(4), byteorder = 'little'))
		msg.Flags = EaInformation(int.from_bytes(buff.read(4), byteorder = 'little'))
		msg.FileId = int.from_bytes(buff.read(16), byteorder = 'little')

		if msg.InputBufferLength > 0:
			buff.seek(msg.InputBufferOffset, io.SEEK_SET)
			msg.Buffer= buff.read(msg.InputBufferLength)

		return msg

	def __repr__(self):
		t = '==== SMB2 QUERY INFO REQ ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'InfoType: %s\r\n' % self.InfoType
		t += 'FileInfoClass: %s\r\n' % self.FileInfoClass
		t += 'OutputBufferLength: %s\r\n' % self.OutputBufferLength
		t += 'InputBufferOffset: %s\r\n' % self.InputBufferOffset
		t += 'InputBufferLength: %s\r\n' % self.InputBufferLength
		t += 'AdditionalInformation: %s\r\n' % self.AdditionalInformation
		t += 'Flags: %s\r\n' % self.Flags
		t += 'FileId: %s\r\n' % self.FileId
		t += 'Buffer: %s\r\n' % self.Buffer
		return t
		
		
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/3b1b3598-a898-44ca-bfac-2dcae065247f
class QUERY_INFO_REPLY:
	def __init__(self):
		self.StructureSize = 9
		self.OutputBufferOffset  = None
		self.OutputBufferLength  = None
		self.Buffer = None
		
		#high-level param
		self.Data = None
		
	def to_bytes(self):
		pass

	@staticmethod
	def from_bytes(bbuff):
		return QUERY_INFO_REPLY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = QUERY_INFO_REPLY()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 9
		msg.OutputBufferOffset  = int.from_bytes(buff.read(2), byteorder='little')
		msg.OutputBufferLength  = int.from_bytes(buff.read(4), byteorder='little')
		

		if msg.OutputBufferLength > 0:
			buff.seek(msg.OutputBufferOffset, io.SEEK_SET)
			msg.Data = buff.read(msg.OutputBufferLength)

		return msg

	def __repr__(self):
		t = '==== SMB2 QUERY INFO REQ ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'OutputBufferOffset: %s\r\n' % self.OutputBufferOffset
		t += 'OutputBufferLength: %s\r\n' % self.OutputBufferLength
		t += 'Buffer: %s\r\n' % self.Buffer
		return t